import re
from _typeshed import ConvertibleToInt
from collections.abc import Callable, Sequence
from datetime import date, datetime, time, timedelta
from typing import Any, Literal

class ValidationError(ValueError): ...

def check_ip(s: str) -> str: ...
def check_positive(s: ConvertibleToInt) -> int: ...
def check_identifier(s: str) -> str: ...

isbn_re: re.Pattern[str]

def isbn10_checksum(digits: Sequence[ConvertibleToInt]) -> str: ...
def isbn13_checksum(digits: Sequence[ConvertibleToInt]) -> str: ...
def check_isbn(s: str, convert_to: Literal[10, 13] | None = None) -> str: ...
def isbn10_to_isbn13(s: str) -> str: ...
def isbn13_to_isbn10(s: str) -> str: ...

email_re: re.Pattern[str]
rfc2822_email_re: re.Pattern[str]

def check_email(s: str) -> str: ...
def check_rfc2822_email(s: str) -> str: ...

date_str_list: list[str]
date_re_list: list[re.Pattern[str]]
time_str: str
time_re: re.Pattern[str]
datetime_re_list: list[re.Pattern[str]]
month_lists: list[list[str]]
month_list: list[str]
i: int
month: str
month_dict: dict[str, int]

def str2date(s: str) -> date: ...
def str2time(s: str) -> time: ...
def str2datetime(s: str) -> datetime: ...
def str2timedelta(s: str) -> timedelta: ...
def timedelta2str(td: timedelta) -> str: ...

converters: dict[type | str, tuple[Callable[[str], Any], type[str], str | None]]  # Any type from types above

def str2py(
    value: str, type: str | type | tuple[Callable[[str], Any], type[str], str | None] | None
) -> Any: ...  # Any type from types above



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
from __future__ import annotations
import asyncio
import enum
import json
import logging
import re
import sys
import uuid
import traceback
from concurrent.futures import Future
from functools import partial
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from pygls.server import LanguageServer, WebSocketTransportAdapter


import attrs
from cattrs.errors import ClassValidationError

from lsprotocol.types import (
    CANCEL_REQUEST,
    EXIT,
    WORKSPACE_EXECUTE_COMMAND,
    ResponseError,
    ResponseErrorMessage,
)

from pygls.exceptions import (
    JsonRpcException,
    JsonRpcInternalError,
    JsonRpcInvalidParams,
    JsonRpcMethodNotFound,
    JsonRpcRequestCancelled,
    FeatureNotificationError,
    FeatureRequestError,
)
from pygls.feature_manager import FeatureManager, is_thread_function

logger = logging.getLogger(__name__)


@attrs.define
class JsonRPCNotification:
    """A class that represents a generic json rpc notification message.
    Used as a fallback for unknown types.
    """

    method: str
    jsonrpc: str
    params: Any


@attrs.define
class JsonRPCRequestMessage:
    """A class that represents a generic json rpc request message.
    Used as a fallback for unknown types.
    """

    id: Union[int, str]
    method: str
    jsonrpc: str
    params: Any


@attrs.define
class JsonRPCResponseMessage:
    """A class that represents a generic json rpc response message.
    Used as a fallback for unknown types.
    """

    id: Union[int, str]
    jsonrpc: str
    result: Any


class JsonRPCProtocol(asyncio.Protocol):
    """Json RPC protocol implementation using on top of `asyncio.Protocol`.

    Specification of the protocol can be found here:
        https://www.jsonrpc.org/specification

    This class provides bidirectional communication which is needed for LSP.
    """

    CHARSET = "utf-8"
    CONTENT_TYPE = "application/vscode-jsonrpc"

    MESSAGE_PATTERN = re.compile(
        rb"^(?:[^\r\n]+\r\n)*"
        + rb"Content-Length: (?P<length>\d+)\r\n"
        + rb"(?:[^\r\n]+\r\n)*\r\n"
        + rb"(?P<body>{.*)",
        re.DOTALL,
    )

    VERSION = "2.0"

    def __init__(self, server: LanguageServer, converter):
        self._server = server
        self._converter = converter

        self._shutdown = False

        # Book keeping for in-flight requests
        self._request_futures: Dict[str, Future[Any]] = {}
        self._result_types: Dict[str, Any] = {}

        self.fm = FeatureManager(server, converter)
        self.transport: Optional[
            Union[asyncio.WriteTransport, WebSocketTransportAdapter]
        ] = None
        self._message_buf: List[bytes] = []

        self._send_only_body = False

    def __call__(self):
        return self

    def _execute_notification(self, handler, *params):
        """Executes notification message handler."""
        if asyncio.iscoroutinefunction(handler):
            future = asyncio.ensure_future(handler(*params))
            future.add_done_callback(self._execute_notification_callback)
        else:
            if is_thread_function(handler):
                self._server.thread_pool.apply_async(handler, (*params,))
            else:
                handler(*params)

    def _execute_notification_callback(self, future):
        """Success callback used for coroutine notification message."""
        if future.exception():
            try:
                raise future.exception()
            except Exception:
                error = JsonRpcInternalError.of(sys.exc_info())
                logger.exception('Exception occurred in notification: "%s"', error)

            # Revisit. Client does not support response with msg_id = None
            # https://stackoverflow.com/questions/31091376/json-rpc-2-0-allow-notifications-to-have-an-error-response
            # self._send_response(None, error=error)

    def _execute_request(self, msg_id, handler, params):
        """Executes request message handler."""

        if asyncio.iscoroutinefunction(handler):
            future = asyncio.ensure_future(handler(params))
            self._request_futures[msg_id] = future
            future.add_done_callback(partial(self._execute_request_callback, msg_id))
        else:
            # Can't be canceled
            if is_thread_function(handler):
                self._server.thread_pool.apply_async(
                    handler,
                    (params,),
                    callback=partial(
                        self._send_response,
                        msg_id,
                    ),
                    error_callback=partial(self._execute_request_err_callback, msg_id),
                )
            else:
                self._send_response(msg_id, handler(params))

    def _execute_request_callback(self, msg_id, future):
        """Success callback used for coroutine request message."""
        try:
            if not future.cancelled():
                self._send_response(msg_id, result=future.result())
            else:
                self._send_response(
                    msg_id,
                    error=JsonRpcRequestCancelled(
                        f'Request with id "{msg_id}" is canceled'
                    ).to_response_error(),
                )
            self._request_futures.pop(msg_id, None)
        except Exception:
            error = JsonRpcInternalError.of(sys.exc_info())
            logger.exception('Exception occurred for message "%s": %s', msg_id, error)
            self._send_response(msg_id, error=error.to_response_error())

    def _execute_request_err_callback(self, msg_id, exc):
        """Error callback used for coroutine request message."""
        exc_info = (type(exc), exc, None)
        error = JsonRpcInternalError.of(exc_info)
        logger.exception('Exception occurred for message "%s": %s', msg_id, error)
        self._send_response(msg_id, error=error.to_response_error())

    def _get_handler(self, feature_name):
        """Returns builtin or used defined feature by name if exists."""
        try:
            return self.fm.builtin_features[feature_name]
        except KeyError:
            try:
                return self.fm.features[feature_name]
            except KeyError:
                raise JsonRpcMethodNotFound.of(feature_name)

    def _handle_cancel_notification(self, msg_id):
        """Handles a cancel notification from the client."""
        future = self._request_futures.pop(msg_id, None)

        if not future:
            logger.warning('Cancel notification for unknown message id "%s"', msg_id)
            return

        # Will only work if the request hasn't started executing
        if future.cancel():
            logger.info('Cancelled request with id "%s"', msg_id)

    def _handle_notification(self, method_name, params):
        """Handles a notification from the client."""
        if method_name == CANCEL_REQUEST:
            self._handle_cancel_notification(params.id)
            return

        try:
            handler = self._get_handler(method_name)
            self._execute_notification(handler, params)
        except (KeyError, JsonRpcMethodNotFound):
            logger.warning('Ignoring notification for unknown method "%s"', method_name)
        except Exception as error:
            logger.exception(
                'Failed to handle notification "%s": %s',
                method_name,
                params,
                exc_info=True,
            )
            self._server._report_server_error(error, FeatureNotificationError)

    def _handle_request(self, msg_id, method_name, params):
        """Handles a request from the client."""
        try:
            handler = self._get_handler(method_name)

            # workspace/executeCommand is a special case
            if method_name == WORKSPACE_EXECUTE_COMMAND:
                handler(params, msg_id)
            else:
                self._execute_request(msg_id, handler, params)

        except JsonRpcException as error:
            logger.exception(
                "Failed to handle request %s %s %s",
                msg_id,
                method_name,
                params,
                exc_info=True,
            )
            self._send_response(msg_id, None, error.to_response_error())
            self._server._report_server_error(error, FeatureRequestError)
        except Exception as error:
            logger.exception(
                "Failed to handle request %s %s %s",
                msg_id,
                method_name,
                params,
                exc_info=True,
            )
            err = JsonRpcInternalError.of(sys.exc_info()).to_response_error()
            self._send_response(msg_id, None, err)
            self._server._report_server_error(error, FeatureRequestError)

    def _handle_response(self, msg_id, result=None, error=None):
        """Handles a response from the client."""
        future = self._request_futures.pop(msg_id, None)

        if not future:
            logger.warning('Received response to unknown message id "%s"', msg_id)
            return

        if error is not None:
            logger.debug('Received error response to message "%s": %s', msg_id, error)
            future.set_exception(JsonRpcException.from_error(error))
        else:
            logger.debug('Received result for message "%s": %s', msg_id, result)
            future.set_result(result)

    def _serialize_message(self, data):
        """Function used to serialize data sent to the client."""

        if hasattr(data, "__attrs_attrs__"):
            return self._converter.unstructure(data)

        if isinstance(data, enum.Enum):
            return data.value

        return data.__dict__

    def _deserialize_message(self, data):
        """Function used to deserialize data recevied from the client."""

        if "jsonrpc" not in data:
            return data

        try:
            if "id" in data:
                if "error" in data:
                    return self._converter.structure(data, ResponseErrorMessage)
                elif "method" in data:
                    request_type = (
                        self.get_message_type(data["method"]) or JsonRPCRequestMessage
                    )
                    return self._converter.structure(data, request_type)
                else:
                    response_type = (
                        self._result_types.pop(data["id"]) or JsonRPCResponseMessage
                    )
                    return self._converter.structure(data, response_type)

            else:
                method = data.get("method", "")
                notification_type = self.get_message_type(method) or JsonRPCNotification
                return self._converter.structure(data, notification_type)

        except ClassValidationError as exc:
            logger.error("Unable to deserialize message\n%s", traceback.format_exc())
            raise JsonRpcInvalidParams() from exc

        except Exception as exc:
            logger.error("Unable to deserialize message\n%s", traceback.format_exc())
            raise JsonRpcInternalError() from exc

    def _procedure_handler(self, message):
        """Delegates message to handlers depending on message type."""

        if message.jsonrpc != JsonRPCProtocol.VERSION:
            logger.warning('Unknown message "%s"', message)
            return

        if self._shutdown and getattr(message, "method", "") != EXIT:
            logger.warning("Server shutting down. No more requests!")
            return

        if hasattr(message, "method"):
            if hasattr(message, "id"):
                logger.debug("Request message received.")
                self._handle_request(message.id, message.method, message.params)
            else:
                logger.debug("Notification message received.")
                self._handle_notification(message.method, message.params)
        else:
            if hasattr(message, "error"):
                logger.debug("Error message received.")
                self._handle_response(message.id, None, message.error)
            else:
                logger.debug("Response message received.")
                self._handle_response(message.id, message.result)

    def _send_data(self, data):
        """Sends data to the client."""
        if not data:
            return

        if self.transport is None:
            logger.error("Unable to send data, no available transport!")
            return

        try:
            body = json.dumps(data, default=self._serialize_message)
            logger.info("Sending data: %s", body)

            if self._send_only_body:
                # Mypy/Pyright seem to think `write()` wants `"bytes | bytearray | memoryview"`
                # But runtime errors with anything but `str`.
                self.transport.write(body)  # type: ignore
                return

            header = (
                f"Content-Length: {len(body)}\r\n"
                f"Content-Type: {self.CONTENT_TYPE}; charset={self.CHARSET}\r\n\r\n"
            ).encode(self.CHARSET)

            self.transport.write(header + body.encode(self.CHARSET))
        except Exception as error:
            logger.exception("Error sending data", exc_info=True)
            self._server._report_server_error(error, JsonRpcInternalError)

    def _send_response(
        self, msg_id, result=None, error: Union[ResponseError, None] = None
    ):
        """Sends a JSON RPC response to the client.

        Args:
            msg_id(str): Id from request
            result(any): Result returned by handler
            error(any): Error returned by handler
        """

        if error is not None:
            response = ResponseErrorMessage(id=msg_id, error=error)

        else:
            response_type = self._result_types.pop(msg_id, JsonRPCResponseMessage)
            response = response_type(
                id=msg_id, result=result, jsonrpc=JsonRPCProtocol.VERSION
            )

        self._send_data(response)

    def connection_lost(self, exc):
        """Method from base class, called when connection is lost, in which case we
        want to shutdown the server's process as well.
        """
        logger.error("Connection to the client is lost! Shutting down the server.")
        sys.exit(1)

    def connection_made(  # type: ignore # see: https://github.com/python/typeshed/issues/3021
        self,
        transport: asyncio.Transport,
    ):
        """Method from base class, called when connection is established"""
        self.transport = transport

    def data_received(self, data: bytes):
        try:
            self._data_received(data)
        except Exception as error:
            logger.exception("Error receiving data", exc_info=True)
            self._server._report_server_error(error, JsonRpcInternalError)

    def _data_received(self, data: bytes):
        """Method from base class, called when server receives the data"""
        logger.debug("Received %r", data)

        while len(data):
            # Append the incoming chunk to the message buffer
            self._message_buf.append(data)

            # Look for the body of the message
            message = b"".join(self._message_buf)
            found = JsonRPCProtocol.MESSAGE_PATTERN.fullmatch(message)

            body = found.group("body") if found else b""
            length = int(found.group("length")) if found else 1

            if len(body) < length:
                # Message is incomplete; bail until more data arrives
                return

            # Message is complete;
            # extract the body and any remaining data,
            # and reset the buffer for the next message
            body, data = body[:length], body[length:]
            self._message_buf = []

            # Parse the body
            self._procedure_handler(
                json.loads(
                    body.decode(self.CHARSET), object_hook=self._deserialize_message
                )
            )

    def get_message_type(self, method: str) -> Optional[Type]:
        """Return the type definition of the message associated with the given method."""
        return None

    def get_result_type(self, method: str) -> Optional[Type]:
        """Return the type definition of the result associated with the given method."""
        return None

    def notify(self, method: str, params=None):
        """Sends a JSON RPC notification to the client."""

        logger.debug("Sending notification: '%s' %s", method, params)

        notification_type = self.get_message_type(method) or JsonRPCNotification
        notification = notification_type(
            method=method, params=params, jsonrpc=JsonRPCProtocol.VERSION
        )

        self._send_data(notification)

    def send_request(self, method, params=None, callback=None, msg_id=None):
        """Sends a JSON RPC request to the client.

        Args:
            method(str): The method name of the message to send
            params(any): The payload of the message

        Returns:
            Future that will be resolved once a response has been received
        """

        if msg_id is None:
            msg_id = str(uuid.uuid4())

        request_type = self.get_message_type(method) or JsonRPCRequestMessage
        logger.debug('Sending request with id "%s": %s %s', msg_id, method, params)

        request = request_type(
            id=msg_id,
            method=method,
            params=params,
            jsonrpc=JsonRPCProtocol.VERSION,
        )

        future = Future()  # type: ignore[var-annotated]
        # If callback function is given, call it when result is received
        if callback:

            def wrapper(future: Future):
                result = future.result()
                logger.info("Client response for %s received: %s", params, result)
                callback(result)

            future.add_done_callback(wrapper)

        self._request_futures[msg_id] = future
        self._result_types[msg_id] = self.get_result_type(method)

        self._send_data(request)

        return future

    def send_request_async(self, method, params=None, msg_id=None):
        """Calls `send_request` and wraps `concurrent.futures.Future` with
        `asyncio.Future` so it can be used with `await` keyword.

        Args:
            method(str): The method name of the message to send
            params(any): The payload of the message
            msg_id(str|int): Optional, message id

        Returns:
            `asyncio.Future` that can be awaited
        """
        return asyncio.wrap_future(
            self.send_request(method, params=params, msg_id=msg_id)
        )

    def thread(self):
        """Decorator that mark function to execute it in a thread."""
        return self.fm.thread()



from _typeshed import StrPath, SupportsWrite
from collections.abc import Collection
from typing_extensions import TypeAlias, TypeVar

from networkx.classes.graph import Graph, _Node

__all__ = ["to_latex_raw", "to_latex", "write_latex"]

# runtime requires a dict but it doesn't mutate it, we use a bounded typevar as
# a values type to make type checkers treat the dict covariantely
_PosT = TypeVar("_PosT", bound=Collection[float] | str)
_Pos: TypeAlias = str | dict[_Node, _PosT]

def to_latex_raw(
    G: Graph[_Node],
    pos: _Pos[_Node, _PosT] = "pos",
    tikz_options: str = "",
    default_node_options: str = "",
    node_options: str | dict[_Node, str] = "node_options",
    node_label: str | dict[_Node, str] = "label",
    default_edge_options: str = "",
    edge_options: str | dict[tuple[_Node, _Node], str] = "edge_options",
    edge_label: str | dict[tuple[_Node, _Node], str] = "label",
    edge_label_options: str | dict[tuple[_Node, _Node], str] = "edge_label_options",
) -> str: ...
def to_latex(
    Gbunch: Graph[_Node] | Collection[Graph[_Node]],
    pos: _Pos[_Node, _PosT] | Collection[_Pos[_Node, _PosT]] = "pos",
    tikz_options: str = "",
    default_node_options: str = "",
    node_options: str | dict[_Node, str] = "node_options",
    node_label: str | dict[_Node, str] = "node_label",
    default_edge_options: str = "",
    edge_options: str | dict[tuple[_Node, _Node], str] = "edge_options",
    edge_label: str | dict[tuple[_Node, _Node], str] = "edge_label",
    edge_label_options: str | dict[tuple[_Node, _Node], str] = "edge_label_options",
    caption: str = "",
    latex_label: str = "",
    sub_captions: Collection[str] | None = None,
    sub_labels: Collection[str] | None = None,
    n_rows: int = 1,
    as_document: bool = True,
    document_wrapper: str = ...,
    figure_wrapper: str = ...,
    subfigure_wrapper: str = ...,
) -> str: ...
def write_latex(
    Gbunch: Graph[_Node] | Collection[Graph[_Node]],
    path: StrPath | SupportsWrite[str],
    *,
    # **options passed to `to_latex`
    pos: _Pos[_Node, _PosT] | Collection[_Pos[_Node, _PosT]] = "pos",
    tikz_options: str = "",
    default_node_options: str = "",
    node_options: str | dict[_Node, str] = "node_options",
    node_label: str | dict[_Node, str] = "node_label",
    default_edge_options: str = "",
    edge_options: str | dict[tuple[_Node, _Node], str] = "edge_options",
    edge_label: str | dict[tuple[_Node, _Node], str] = "edge_label",
    edge_label_options: str | dict[tuple[_Node, _Node], str] = "edge_label_options",
    caption: str = "",
    latex_label: str = "",
    sub_captions: Collection[str] | None = None,
    sub_labels: Collection[str] | None = None,
    n_rows: int = 1,
    as_document: bool = True,
    document_wrapper: str = ...,
    figure_wrapper: str = ...,
    subfigure_wrapper: str = ...,
) -> None: ...



from _typeshed import ConvertibleToInt, Incomplete
from typing import ClassVar, Literal

from openpyxl.descriptors.base import Alias, Integer
from openpyxl.descriptors.serialisable import Serialisable

class SheetBackgroundPicture(Serialisable):
    tagname: ClassVar[str]
    id: Incomplete
    def __init__(self, id) -> None: ...

class DrawingHF(Serialisable):
    id: Incomplete
    lho: Integer[Literal[True]]
    leftHeaderOddPages: Alias
    lhe: Integer[Literal[True]]
    leftHeaderEvenPages: Alias
    lhf: Integer[Literal[True]]
    leftHeaderFirstPage: Alias
    cho: Integer[Literal[True]]
    centerHeaderOddPages: Alias
    che: Integer[Literal[True]]
    centerHeaderEvenPages: Alias
    chf: Integer[Literal[True]]
    centerHeaderFirstPage: Alias
    rho: Integer[Literal[True]]
    rightHeaderOddPages: Alias
    rhe: Integer[Literal[True]]
    rightHeaderEvenPages: Alias
    rhf: Integer[Literal[True]]
    rightHeaderFirstPage: Alias
    lfo: Integer[Literal[True]]
    leftFooterOddPages: Alias
    lfe: Integer[Literal[True]]
    leftFooterEvenPages: Alias
    lff: Integer[Literal[True]]
    leftFooterFirstPage: Alias
    cfo: Integer[Literal[True]]
    centerFooterOddPages: Alias
    cfe: Integer[Literal[True]]
    centerFooterEvenPages: Alias
    cff: Integer[Literal[True]]
    centerFooterFirstPage: Alias
    rfo: Integer[Literal[True]]
    rightFooterOddPages: Alias
    rfe: Integer[Literal[True]]
    rightFooterEvenPages: Alias
    rff: Integer[Literal[True]]
    rightFooterFirstPage: Alias
    def __init__(
        self,
        id=None,
        lho: ConvertibleToInt | None = None,
        lhe: ConvertibleToInt | None = None,
        lhf: ConvertibleToInt | None = None,
        cho: ConvertibleToInt | None = None,
        che: ConvertibleToInt | None = None,
        chf: ConvertibleToInt | None = None,
        rho: ConvertibleToInt | None = None,
        rhe: ConvertibleToInt | None = None,
        rhf: ConvertibleToInt | None = None,
        lfo: ConvertibleToInt | None = None,
        lfe: ConvertibleToInt | None = None,
        lff: ConvertibleToInt | None = None,
        cfo: ConvertibleToInt | None = None,
        cfe: ConvertibleToInt | None = None,
        cff: ConvertibleToInt | None = None,
        rfo: ConvertibleToInt | None = None,
        rfe: ConvertibleToInt | None = None,
        rff: ConvertibleToInt | None = None,
    ) -> None: ...



from _typeshed import FileDescriptor
from collections.abc import Callable, Collection, Iterable
from types import TracebackType
from typing import Any, Generic, Protocol, TypeVar, overload, type_check_only
from typing_extensions import Self, TypeVarTuple, Unpack, disjoint_base

from gevent._greenlet_primitives import SwitchOutGreenletWithLoop
from gevent._types import _Loop, _Watcher
from gevent.hub import Hub
from gevent.socket import socket

__all__ = ["WaitOperationsGreenlet", "iwait_on_objects", "wait_on_objects", "wait_read", "wait_write", "wait_readwrite"]

_T = TypeVar("_T")
_Ts = TypeVarTuple("_Ts")
_WaitableT = TypeVar("_WaitableT", bound=_Waitable)

@type_check_only
class _Waitable(Protocol):
    def rawlink(self, callback: Callable[[Any], object], /) -> object: ...
    def unlink(self, callback: Callable[[Any], object], /) -> object: ...

class WaitOperationsGreenlet(SwitchOutGreenletWithLoop):
    loop: _Loop
    def wait(self, watcher: _Watcher) -> None: ...
    def cancel_waits_close_and_then(
        self,
        watchers: Iterable[_Watcher],
        exc_kind: type[BaseException] | BaseException,
        then: Callable[[Unpack[_Ts]], object],
        *then_args: Unpack[_Ts],
    ) -> None: ...
    def cancel_wait(self, watcher: _Watcher, error: type[BaseException] | BaseException, close_watcher: bool = False) -> None: ...

@disjoint_base
class _WaitIterator(Generic[_T]):
    def __init__(self, objects: Collection[_T], hub: Hub, timeout: float, count: None | int) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, typ: type[BaseException] | None, value: BaseException | None, tb: TracebackType | None) -> None: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> _T: ...
    next = __next__

@overload
def iwait_on_objects(objects: None, timeout: float | None = None, count: int | None = None) -> list[bool]: ...
@overload
def iwait_on_objects(
    objects: Collection[_WaitableT], timeout: float | None = None, count: int | None = None
) -> _WaitIterator[_WaitableT]: ...
@overload
def wait_on_objects(objects: None = None, timeout: float | None = None, count: int | None = None) -> bool: ...
@overload
def wait_on_objects(
    objects: Collection[_WaitableT], timeout: float | None = None, count: int | None = None
) -> list[_WaitableT]: ...
def set_default_timeout_error(e: type[BaseException]) -> None: ...
def wait_on_socket(socket: socket, watcher: _Watcher, timeout_exc: type[BaseException] | BaseException | None = None) -> None: ...
def wait_on_watcher(
    watcher: _Watcher,
    timeout: float | None = None,
    timeout_exc: type[BaseException] | BaseException = ...,
    hub: Hub | None = None,
) -> None: ...
def wait_read(
    fileno: FileDescriptor, timeout: float | None = None, timeout_exc: type[BaseException] | BaseException = ...
) -> None: ...
def wait_write(
    fileno: FileDescriptor, timeout: float | None = None, timeout_exc: type[BaseException] | BaseException = ...
) -> None: ...
def wait_readwrite(
    fileno: FileDescriptor, timeout: float | None = None, timeout_exc: type[BaseException] | BaseException = ...
) -> None: ...



import io
from _typeshed import Incomplete
from collections.abc import Generator

from pyasn1 import error as error
from pyasn1.type import univ as univ

class CachingStreamWrapper(io.IOBase):
    def __init__(self, raw) -> None: ...
    def peek(self, n): ...
    def seekable(self): ...
    def seek(self, n: int = -1, whence=0): ...
    def read(self, n: int = -1): ...
    @property
    def markedPosition(self): ...
    def tell(self): ...

def asSeekableStream(substrate): ...
def isEndOfStream(substrate) -> Generator[Incomplete, None, None]: ...
def peekIntoStream(substrate, size: int = -1) -> Generator[Incomplete, None, None]: ...
def readFromStream(substrate, size: int = -1, context=None) -> Generator[Incomplete, None, None]: ...



from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["degree_centrality", "in_degree_centrality", "out_degree_centrality"]

@_dispatchable
def degree_centrality(G: Graph[_Node]) -> dict[_Node, float]: ...
@_dispatchable
def in_degree_centrality(G: Graph[_Node]) -> dict[_Node, float]: ...
@_dispatchable
def out_degree_centrality(G: Graph[_Node]) -> dict[_Node, float]: ...



import datetime
from _typeshed import Incomplete
from collections.abc import Iterable
from datetime import timedelta, tzinfo
from typing import Final, Literal, overload

from .base import Component
from .behavior import Behavior

DATENAMES: Final[tuple[str, ...]]
RULENAMES: Final[tuple[str, ...]]
DATESANDRULES: Final[tuple[str, ...]]
PRODID: Final[str]
WEEKDAYS: Final[tuple[str, ...]]
FREQUENCIES: Final[tuple[str, ...]]
ZERO_DELTA: Final[timedelta]
twoHours: Final[timedelta]

@overload
def toUnicode(s: None) -> None: ...
@overload
def toUnicode(s: str | bytes) -> str: ...
def registerTzid(tzid: str | bytes, tzinfo) -> None: ...
def getTzid(tzid: str, smart: bool = True): ...

utc: tzinfo  # dateutil.tz.tz.tzutc (subclass of tzinfo)

class TimezoneComponent(Component):
    isNative: bool
    behavior: Incomplete
    tzinfo: Incomplete
    name: str
    useBegin: bool
    def __init__(self, tzinfo=None, *args, **kwds) -> None: ...
    @classmethod
    def registerTzinfo(cls, tzinfo) -> str | None: ...
    def gettzinfo(self): ...
    tzid: Incomplete
    daylight: Incomplete
    standard: Incomplete
    def settzinfo(self, tzinfo, start: int = 2000, end: int = 2030) -> None: ...
    normal_attributes: Incomplete
    @staticmethod
    def pickTzid(tzinfo, allowUTC: bool = False) -> str | None: ...
    def prettyPrint(self, level: int, tabwidth: int) -> None: ...  # type: ignore[override]

class RecurringComponent(Component):
    isNative: bool
    def __init__(self, *args, **kwds) -> None: ...
    def getrruleset(self, addRDate: bool = False): ...
    def setrruleset(self, rruleset): ...
    rruleset: Incomplete
    def __setattr__(self, name, value) -> None: ...

class TextBehavior(Behavior):
    base64string: str
    @classmethod
    def decode(cls, line) -> None: ...
    @classmethod
    def encode(cls, line) -> None: ...

class VCalendarComponentBehavior(Behavior):
    defaultBehavior: Incomplete
    isComponent: bool

class RecurringBehavior(VCalendarComponentBehavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...
    @staticmethod
    def generateImplicitParameters(obj) -> None: ...

class DateTimeBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @classmethod
    def transformFromNative(cls, obj): ...

class UTCDateTimeBehavior(DateTimeBehavior):
    forceUTC: bool

class DateOrDateTimeBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class MultiDateBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class MultiTextBehavior(Behavior):
    listSeparator: str
    @classmethod
    def decode(cls, line) -> None: ...
    @classmethod
    def encode(cls, line) -> None: ...

class SemicolonMultiTextBehavior(MultiTextBehavior):
    listSeparator: str

class VCalendar2_0(VCalendarComponentBehavior):
    name: str
    description: str
    versionString: str
    sortFirst: Incomplete
    @classmethod
    def generateImplicitParameters(cls, obj) -> None: ...
    @classmethod
    def serialize(cls, obj, buf, lineLength, validate: bool = True): ...

class VTimezone(VCalendarComponentBehavior):
    name: str
    hasNative: bool
    description: str
    sortFirst: Incomplete
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class TZID(Behavior): ...

class DaylightOrStandard(VCalendarComponentBehavior):
    hasNative: bool

class VEvent(RecurringBehavior):
    name: str
    sortFirst: Incomplete
    description: str
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]

class VTodo(RecurringBehavior):
    name: str
    description: str
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]

class VJournal(RecurringBehavior):
    name: str

class VFreeBusy(VCalendarComponentBehavior):
    name: str
    description: str
    sortFirst: Incomplete

class VAlarm(VCalendarComponentBehavior):
    name: str
    description: str
    @staticmethod
    def generateImplicitParameters(obj) -> None: ...
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]

class VAvailability(VCalendarComponentBehavior):
    name: str
    description: str
    sortFirst: Incomplete
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]

class Available(RecurringBehavior):
    name: str
    sortFirst: Incomplete
    description: str
    @classmethod
    def validate(cls, obj, raiseException: bool, *args) -> bool: ...  # type: ignore[override]

class Duration(Behavior):
    name: str
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class Trigger(Behavior):
    name: str
    description: str
    hasNative: bool
    forceUTC: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class PeriodBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @classmethod
    def transformFromNative(cls, obj): ...

class FreeBusy(PeriodBehavior):
    name: str
    forceUTC: bool

class RRule(Behavior): ...

utcDateTimeList: list[str]
dateTimeOrDateList: list[str]
textList: list[str]

def numToDigits(num: float | None, places: int) -> str: ...
def timedeltaToString(delta: datetime.timedelta) -> str: ...
def timeToString(dateOrDateTime: datetime.date | datetime.datetime) -> str: ...
def dateToString(date: datetime.date) -> str: ...
def dateTimeToString(dateTime: datetime.datetime, convertToUTC: bool = False) -> str: ...
def deltaToOffset(delta: datetime.timedelta) -> str: ...
def periodToString(period, convertToUTC: bool = False): ...
def isDuration(s: str) -> bool: ...
def stringToDate(s: str) -> datetime.date: ...
def stringToDateTime(s: str, tzinfo: datetime._TzInfo | None = None, strict: bool = False) -> datetime.datetime: ...

escapableCharList: str

def stringToTextValues(
    s: str, listSeparator: str = ",", charList: Iterable[str] | None = None, strict: bool = False
) -> list[str]: ...
def stringToDurations(s: str, strict: bool = False) -> list[datetime.timedelta]: ...
def parseDtstart(contentline, allowSignatureMismatch: bool = False): ...
def stringToPeriod(s: str, tzinfo=None): ...
def getTransition(transitionTo: Literal["daylight", "standard"], year: int, tzinfo: datetime._TzInfo): ...
def tzinfo_eq(
    tzinfo1: datetime._TzInfo | None, tzinfo2: datetime._TzInfo | None, startYear: int = 2000, endYear: int = 2020
) -> bool: ...



from _typeshed import Incomplete

from authlib.oauth2.rfc6750 import BearerTokenValidator

class JWTBearerTokenValidator(BearerTokenValidator):
    issuer: Incomplete
    resource_server: Incomplete
    def __init__(self, issuer, resource_server, *args, **kwargs) -> None: ...
    def get_jwks(self) -> None: ...
    def validate_iss(self, claims, iss: str) -> bool: ...
    def authenticate_token(self, token_string): ...
    def validate_token(self, token, scopes, request, groups=None, roles=None, entitlements=None) -> None: ...



# GENERATED FROM scripts/gen-client.py -- DO NOT EDIT
# flake8: noqa
from concurrent.futures import Future
from lsprotocol import types
from pygls.client import JsonRPCClient
from pygls.protocol import LanguageServerProtocol
from pygls.protocol import default_converter
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union


class BaseLanguageClient(JsonRPCClient):

    def __init__(
        self,
        name: str,
        version: str,
        protocol_cls=LanguageServerProtocol,
        converter_factory=default_converter,
        **kwargs,
    ):
        self.name = name
        self.version = version
        super().__init__(protocol_cls, converter_factory, **kwargs)

    def call_hierarchy_incoming_calls(
        self,
        params: types.CallHierarchyIncomingCallsParams,
        callback: Optional[Callable[[Optional[List[types.CallHierarchyIncomingCall]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`callHierarchy/incomingCalls` request.

        A request to resolve the incoming calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("callHierarchy/incomingCalls", params, callback)

    async def call_hierarchy_incoming_calls_async(
        self,
        params: types.CallHierarchyIncomingCallsParams,
    ) -> Optional[List[types.CallHierarchyIncomingCall]]:
        """Make a :lsp:`callHierarchy/incomingCalls` request.

        A request to resolve the incoming calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("callHierarchy/incomingCalls", params)

    def call_hierarchy_outgoing_calls(
        self,
        params: types.CallHierarchyOutgoingCallsParams,
        callback: Optional[Callable[[Optional[List[types.CallHierarchyOutgoingCall]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`callHierarchy/outgoingCalls` request.

        A request to resolve the outgoing calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("callHierarchy/outgoingCalls", params, callback)

    async def call_hierarchy_outgoing_calls_async(
        self,
        params: types.CallHierarchyOutgoingCallsParams,
    ) -> Optional[List[types.CallHierarchyOutgoingCall]]:
        """Make a :lsp:`callHierarchy/outgoingCalls` request.

        A request to resolve the outgoing calls for a given `CallHierarchyItem`.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("callHierarchy/outgoingCalls", params)

    def code_action_resolve(
        self,
        params: types.CodeAction,
        callback: Optional[Callable[[types.CodeAction], None]] = None,
    ) -> Future:
        """Make a :lsp:`codeAction/resolve` request.

        Request to resolve additional information for a given code action.The request's
        parameter is of type {@link CodeAction} the response
        is of type {@link CodeAction} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("codeAction/resolve", params, callback)

    async def code_action_resolve_async(
        self,
        params: types.CodeAction,
    ) -> types.CodeAction:
        """Make a :lsp:`codeAction/resolve` request.

        Request to resolve additional information for a given code action.The request's
        parameter is of type {@link CodeAction} the response
        is of type {@link CodeAction} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("codeAction/resolve", params)

    def code_lens_resolve(
        self,
        params: types.CodeLens,
        callback: Optional[Callable[[types.CodeLens], None]] = None,
    ) -> Future:
        """Make a :lsp:`codeLens/resolve` request.

        A request to resolve a command for a given code lens.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("codeLens/resolve", params, callback)

    async def code_lens_resolve_async(
        self,
        params: types.CodeLens,
    ) -> types.CodeLens:
        """Make a :lsp:`codeLens/resolve` request.

        A request to resolve a command for a given code lens.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("codeLens/resolve", params)

    def completion_item_resolve(
        self,
        params: types.CompletionItem,
        callback: Optional[Callable[[types.CompletionItem], None]] = None,
    ) -> Future:
        """Make a :lsp:`completionItem/resolve` request.

        Request to resolve additional information for a given completion item.The request's
        parameter is of type {@link CompletionItem} the response
        is of type {@link CompletionItem} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("completionItem/resolve", params, callback)

    async def completion_item_resolve_async(
        self,
        params: types.CompletionItem,
    ) -> types.CompletionItem:
        """Make a :lsp:`completionItem/resolve` request.

        Request to resolve additional information for a given completion item.The request's
        parameter is of type {@link CompletionItem} the response
        is of type {@link CompletionItem} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("completionItem/resolve", params)

    def document_link_resolve(
        self,
        params: types.DocumentLink,
        callback: Optional[Callable[[types.DocumentLink], None]] = None,
    ) -> Future:
        """Make a :lsp:`documentLink/resolve` request.

        Request to resolve additional information for a given document link. The request's
        parameter is of type {@link DocumentLink} the response
        is of type {@link DocumentLink} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("documentLink/resolve", params, callback)

    async def document_link_resolve_async(
        self,
        params: types.DocumentLink,
    ) -> types.DocumentLink:
        """Make a :lsp:`documentLink/resolve` request.

        Request to resolve additional information for a given document link. The request's
        parameter is of type {@link DocumentLink} the response
        is of type {@link DocumentLink} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("documentLink/resolve", params)

    def initialize(
        self,
        params: types.InitializeParams,
        callback: Optional[Callable[[types.InitializeResult], None]] = None,
    ) -> Future:
        """Make a :lsp:`initialize` request.

        The initialize request is sent from the client to the server.
        It is sent once as the request after starting up the server.
        The requests parameter is of type {@link InitializeParams}
        the response if of type {@link InitializeResult} of a Thenable that
        resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("initialize", params, callback)

    async def initialize_async(
        self,
        params: types.InitializeParams,
    ) -> types.InitializeResult:
        """Make a :lsp:`initialize` request.

        The initialize request is sent from the client to the server.
        It is sent once as the request after starting up the server.
        The requests parameter is of type {@link InitializeParams}
        the response if of type {@link InitializeResult} of a Thenable that
        resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("initialize", params)

    def inlay_hint_resolve(
        self,
        params: types.InlayHint,
        callback: Optional[Callable[[types.InlayHint], None]] = None,
    ) -> Future:
        """Make a :lsp:`inlayHint/resolve` request.

        A request to resolve additional properties for an inlay hint.
        The request's parameter is of type {@link InlayHint}, the response is
        of type {@link InlayHint} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("inlayHint/resolve", params, callback)

    async def inlay_hint_resolve_async(
        self,
        params: types.InlayHint,
    ) -> types.InlayHint:
        """Make a :lsp:`inlayHint/resolve` request.

        A request to resolve additional properties for an inlay hint.
        The request's parameter is of type {@link InlayHint}, the response is
        of type {@link InlayHint} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("inlayHint/resolve", params)

    def shutdown(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future:
        """Make a :lsp:`shutdown` request.

        A shutdown request is sent from the client to the server.
        It is sent once when the client decides to shutdown the
        server. The only notification that is sent after a shutdown request
        is the exit event.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("shutdown", params, callback)

    async def shutdown_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`shutdown` request.

        A shutdown request is sent from the client to the server.
        It is sent once when the client decides to shutdown the
        server. The only notification that is sent after a shutdown request
        is the exit event.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("shutdown", params)

    def text_document_code_action(
        self,
        params: types.CodeActionParams,
        callback: Optional[Callable[[Optional[List[Union[types.Command, types.CodeAction]]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/codeAction` request.

        A request to provide commands for the given text document and range.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/codeAction", params, callback)

    async def text_document_code_action_async(
        self,
        params: types.CodeActionParams,
    ) -> Optional[List[Union[types.Command, types.CodeAction]]]:
        """Make a :lsp:`textDocument/codeAction` request.

        A request to provide commands for the given text document and range.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/codeAction", params)

    def text_document_code_lens(
        self,
        params: types.CodeLensParams,
        callback: Optional[Callable[[Optional[List[types.CodeLens]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/codeLens` request.

        A request to provide code lens for the given text document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/codeLens", params, callback)

    async def text_document_code_lens_async(
        self,
        params: types.CodeLensParams,
    ) -> Optional[List[types.CodeLens]]:
        """Make a :lsp:`textDocument/codeLens` request.

        A request to provide code lens for the given text document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/codeLens", params)

    def text_document_color_presentation(
        self,
        params: types.ColorPresentationParams,
        callback: Optional[Callable[[List[types.ColorPresentation]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/colorPresentation` request.

        A request to list all presentation for a color. The request's
        parameter is of type {@link ColorPresentationParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/colorPresentation", params, callback)

    async def text_document_color_presentation_async(
        self,
        params: types.ColorPresentationParams,
    ) -> List[types.ColorPresentation]:
        """Make a :lsp:`textDocument/colorPresentation` request.

        A request to list all presentation for a color. The request's
        parameter is of type {@link ColorPresentationParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/colorPresentation", params)

    def text_document_completion(
        self,
        params: types.CompletionParams,
        callback: Optional[Callable[[Union[List[types.CompletionItem], types.CompletionList, None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/completion` request.

        Request to request completion at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response
        is of type {@link CompletionItem CompletionItem[]} or {@link CompletionList}
        or a Thenable that resolves to such.

        The request can delay the computation of the {@link CompletionItem.detail `detail`}
        and {@link CompletionItem.documentation `documentation`} properties to the `completionItem/resolve`
        request. However, properties that are needed for the initial sorting and filtering, like `sortText`,
        `filterText`, `insertText`, and `textEdit`, must not be changed during resolve.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/completion", params, callback)

    async def text_document_completion_async(
        self,
        params: types.CompletionParams,
    ) -> Union[List[types.CompletionItem], types.CompletionList, None]:
        """Make a :lsp:`textDocument/completion` request.

        Request to request completion at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response
        is of type {@link CompletionItem CompletionItem[]} or {@link CompletionList}
        or a Thenable that resolves to such.

        The request can delay the computation of the {@link CompletionItem.detail `detail`}
        and {@link CompletionItem.documentation `documentation`} properties to the `completionItem/resolve`
        request. However, properties that are needed for the initial sorting and filtering, like `sortText`,
        `filterText`, `insertText`, and `textEdit`, must not be changed during resolve.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/completion", params)

    def text_document_declaration(
        self,
        params: types.DeclarationParams,
        callback: Optional[Callable[[Union[types.Location, List[types.Location], List[types.LocationLink], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/declaration` request.

        A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Declaration} or a typed array of {@link DeclarationLink}
        or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/declaration", params, callback)

    async def text_document_declaration_async(
        self,
        params: types.DeclarationParams,
    ) -> Union[types.Location, List[types.Location], List[types.LocationLink], None]:
        """Make a :lsp:`textDocument/declaration` request.

        A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Declaration} or a typed array of {@link DeclarationLink}
        or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/declaration", params)

    def text_document_definition(
        self,
        params: types.DefinitionParams,
        callback: Optional[Callable[[Union[types.Location, List[types.Location], List[types.LocationLink], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/definition` request.

        A request to resolve the definition location of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPosition}
        the response is of either type {@link Definition} or a typed array of
        {@link DefinitionLink} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/definition", params, callback)

    async def text_document_definition_async(
        self,
        params: types.DefinitionParams,
    ) -> Union[types.Location, List[types.Location], List[types.LocationLink], None]:
        """Make a :lsp:`textDocument/definition` request.

        A request to resolve the definition location of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPosition}
        the response is of either type {@link Definition} or a typed array of
        {@link DefinitionLink} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/definition", params)

    def text_document_diagnostic(
        self,
        params: types.DocumentDiagnosticParams,
        callback: Optional[Callable[[Union[types.RelatedFullDocumentDiagnosticReport, types.RelatedUnchangedDocumentDiagnosticReport]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/diagnostic` request.

        The document diagnostic request definition.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/diagnostic", params, callback)

    async def text_document_diagnostic_async(
        self,
        params: types.DocumentDiagnosticParams,
    ) -> Union[types.RelatedFullDocumentDiagnosticReport, types.RelatedUnchangedDocumentDiagnosticReport]:
        """Make a :lsp:`textDocument/diagnostic` request.

        The document diagnostic request definition.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/diagnostic", params)

    def text_document_document_color(
        self,
        params: types.DocumentColorParams,
        callback: Optional[Callable[[List[types.ColorInformation]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/documentColor` request.

        A request to list all color symbols found in a given text document. The request's
        parameter is of type {@link DocumentColorParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/documentColor", params, callback)

    async def text_document_document_color_async(
        self,
        params: types.DocumentColorParams,
    ) -> List[types.ColorInformation]:
        """Make a :lsp:`textDocument/documentColor` request.

        A request to list all color symbols found in a given text document. The request's
        parameter is of type {@link DocumentColorParams} the
        response is of type {@link ColorInformation ColorInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/documentColor", params)

    def text_document_document_highlight(
        self,
        params: types.DocumentHighlightParams,
        callback: Optional[Callable[[Optional[List[types.DocumentHighlight]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/documentHighlight` request.

        Request to resolve a {@link DocumentHighlight} for a given
        text document position. The request's parameter is of type {@link TextDocumentPosition}
        the request response is an array of type {@link DocumentHighlight}
        or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/documentHighlight", params, callback)

    async def text_document_document_highlight_async(
        self,
        params: types.DocumentHighlightParams,
    ) -> Optional[List[types.DocumentHighlight]]:
        """Make a :lsp:`textDocument/documentHighlight` request.

        Request to resolve a {@link DocumentHighlight} for a given
        text document position. The request's parameter is of type {@link TextDocumentPosition}
        the request response is an array of type {@link DocumentHighlight}
        or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/documentHighlight", params)

    def text_document_document_link(
        self,
        params: types.DocumentLinkParams,
        callback: Optional[Callable[[Optional[List[types.DocumentLink]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/documentLink` request.

        A request to provide document links
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/documentLink", params, callback)

    async def text_document_document_link_async(
        self,
        params: types.DocumentLinkParams,
    ) -> Optional[List[types.DocumentLink]]:
        """Make a :lsp:`textDocument/documentLink` request.

        A request to provide document links
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/documentLink", params)

    def text_document_document_symbol(
        self,
        params: types.DocumentSymbolParams,
        callback: Optional[Callable[[Union[List[types.SymbolInformation], List[types.DocumentSymbol], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/documentSymbol` request.

        A request to list all symbols found in a given text document. The request's
        parameter is of type {@link TextDocumentIdentifier} the
        response is of type {@link SymbolInformation SymbolInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/documentSymbol", params, callback)

    async def text_document_document_symbol_async(
        self,
        params: types.DocumentSymbolParams,
    ) -> Union[List[types.SymbolInformation], List[types.DocumentSymbol], None]:
        """Make a :lsp:`textDocument/documentSymbol` request.

        A request to list all symbols found in a given text document. The request's
        parameter is of type {@link TextDocumentIdentifier} the
        response is of type {@link SymbolInformation SymbolInformation[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/documentSymbol", params)

    def text_document_folding_range(
        self,
        params: types.FoldingRangeParams,
        callback: Optional[Callable[[Optional[List[types.FoldingRange]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/foldingRange` request.

        A request to provide folding ranges in a document. The request's
        parameter is of type {@link FoldingRangeParams}, the
        response is of type {@link FoldingRangeList} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/foldingRange", params, callback)

    async def text_document_folding_range_async(
        self,
        params: types.FoldingRangeParams,
    ) -> Optional[List[types.FoldingRange]]:
        """Make a :lsp:`textDocument/foldingRange` request.

        A request to provide folding ranges in a document. The request's
        parameter is of type {@link FoldingRangeParams}, the
        response is of type {@link FoldingRangeList} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/foldingRange", params)

    def text_document_formatting(
        self,
        params: types.DocumentFormattingParams,
        callback: Optional[Callable[[Optional[List[types.TextEdit]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/formatting` request.

        A request to format a whole document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/formatting", params, callback)

    async def text_document_formatting_async(
        self,
        params: types.DocumentFormattingParams,
    ) -> Optional[List[types.TextEdit]]:
        """Make a :lsp:`textDocument/formatting` request.

        A request to format a whole document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/formatting", params)

    def text_document_hover(
        self,
        params: types.HoverParams,
        callback: Optional[Callable[[Optional[types.Hover]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/hover` request.

        Request to request hover information at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response is of
        type {@link Hover} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/hover", params, callback)

    async def text_document_hover_async(
        self,
        params: types.HoverParams,
    ) -> Optional[types.Hover]:
        """Make a :lsp:`textDocument/hover` request.

        Request to request hover information at a given text document position. The request's
        parameter is of type {@link TextDocumentPosition} the response is of
        type {@link Hover} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/hover", params)

    def text_document_implementation(
        self,
        params: types.ImplementationParams,
        callback: Optional[Callable[[Union[types.Location, List[types.Location], List[types.LocationLink], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/implementation` request.

        A request to resolve the implementation locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Definition} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/implementation", params, callback)

    async def text_document_implementation_async(
        self,
        params: types.ImplementationParams,
    ) -> Union[types.Location, List[types.Location], List[types.LocationLink], None]:
        """Make a :lsp:`textDocument/implementation` request.

        A request to resolve the implementation locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Definition} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/implementation", params)

    def text_document_inlay_hint(
        self,
        params: types.InlayHintParams,
        callback: Optional[Callable[[Optional[List[types.InlayHint]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/inlayHint` request.

        A request to provide inlay hints in a document. The request's parameter is of
        type {@link InlayHintsParams}, the response is of type
        {@link InlayHint InlayHint[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/inlayHint", params, callback)

    async def text_document_inlay_hint_async(
        self,
        params: types.InlayHintParams,
    ) -> Optional[List[types.InlayHint]]:
        """Make a :lsp:`textDocument/inlayHint` request.

        A request to provide inlay hints in a document. The request's parameter is of
        type {@link InlayHintsParams}, the response is of type
        {@link InlayHint InlayHint[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/inlayHint", params)

    def text_document_inline_completion(
        self,
        params: types.InlineCompletionParams,
        callback: Optional[Callable[[Union[types.InlineCompletionList, List[types.InlineCompletionItem], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/inlineCompletion` request.

        A request to provide inline completions in a document. The request's parameter is of
        type {@link InlineCompletionParams}, the response is of type
        {@link InlineCompletion InlineCompletion[]} or a Thenable that resolves to such.

        @since 3.18.0
        @proposed
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/inlineCompletion", params, callback)

    async def text_document_inline_completion_async(
        self,
        params: types.InlineCompletionParams,
    ) -> Union[types.InlineCompletionList, List[types.InlineCompletionItem], None]:
        """Make a :lsp:`textDocument/inlineCompletion` request.

        A request to provide inline completions in a document. The request's parameter is of
        type {@link InlineCompletionParams}, the response is of type
        {@link InlineCompletion InlineCompletion[]} or a Thenable that resolves to such.

        @since 3.18.0
        @proposed
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/inlineCompletion", params)

    def text_document_inline_value(
        self,
        params: types.InlineValueParams,
        callback: Optional[Callable[[Optional[List[Union[types.InlineValueText, types.InlineValueVariableLookup, types.InlineValueEvaluatableExpression]]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/inlineValue` request.

        A request to provide inline values in a document. The request's parameter is of
        type {@link InlineValueParams}, the response is of type
        {@link InlineValue InlineValue[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/inlineValue", params, callback)

    async def text_document_inline_value_async(
        self,
        params: types.InlineValueParams,
    ) -> Optional[List[Union[types.InlineValueText, types.InlineValueVariableLookup, types.InlineValueEvaluatableExpression]]]:
        """Make a :lsp:`textDocument/inlineValue` request.

        A request to provide inline values in a document. The request's parameter is of
        type {@link InlineValueParams}, the response is of type
        {@link InlineValue InlineValue[]} or a Thenable that resolves to such.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/inlineValue", params)

    def text_document_linked_editing_range(
        self,
        params: types.LinkedEditingRangeParams,
        callback: Optional[Callable[[Optional[types.LinkedEditingRanges]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/linkedEditingRange` request.

        A request to provide ranges that can be edited together.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/linkedEditingRange", params, callback)

    async def text_document_linked_editing_range_async(
        self,
        params: types.LinkedEditingRangeParams,
    ) -> Optional[types.LinkedEditingRanges]:
        """Make a :lsp:`textDocument/linkedEditingRange` request.

        A request to provide ranges that can be edited together.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/linkedEditingRange", params)

    def text_document_moniker(
        self,
        params: types.MonikerParams,
        callback: Optional[Callable[[Optional[List[types.Moniker]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/moniker` request.

        A request to get the moniker of a symbol at a given text document position.
        The request parameter is of type {@link TextDocumentPositionParams}.
        The response is of type {@link Moniker Moniker[]} or `null`.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/moniker", params, callback)

    async def text_document_moniker_async(
        self,
        params: types.MonikerParams,
    ) -> Optional[List[types.Moniker]]:
        """Make a :lsp:`textDocument/moniker` request.

        A request to get the moniker of a symbol at a given text document position.
        The request parameter is of type {@link TextDocumentPositionParams}.
        The response is of type {@link Moniker Moniker[]} or `null`.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/moniker", params)

    def text_document_on_type_formatting(
        self,
        params: types.DocumentOnTypeFormattingParams,
        callback: Optional[Callable[[Optional[List[types.TextEdit]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/onTypeFormatting` request.

        A request to format a document on type.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/onTypeFormatting", params, callback)

    async def text_document_on_type_formatting_async(
        self,
        params: types.DocumentOnTypeFormattingParams,
    ) -> Optional[List[types.TextEdit]]:
        """Make a :lsp:`textDocument/onTypeFormatting` request.

        A request to format a document on type.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/onTypeFormatting", params)

    def text_document_prepare_call_hierarchy(
        self,
        params: types.CallHierarchyPrepareParams,
        callback: Optional[Callable[[Optional[List[types.CallHierarchyItem]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/prepareCallHierarchy` request.

        A request to result a `CallHierarchyItem` in a document at a given position.
        Can be used as an input to an incoming or outgoing call hierarchy.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/prepareCallHierarchy", params, callback)

    async def text_document_prepare_call_hierarchy_async(
        self,
        params: types.CallHierarchyPrepareParams,
    ) -> Optional[List[types.CallHierarchyItem]]:
        """Make a :lsp:`textDocument/prepareCallHierarchy` request.

        A request to result a `CallHierarchyItem` in a document at a given position.
        Can be used as an input to an incoming or outgoing call hierarchy.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/prepareCallHierarchy", params)

    def text_document_prepare_rename(
        self,
        params: types.PrepareRenameParams,
        callback: Optional[Callable[[Union[types.Range, types.PrepareRenameResult_Type1, types.PrepareRenameResult_Type2, None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/prepareRename` request.

        A request to test and perform the setup necessary for a rename.

        @since 3.16 - support for default behavior
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/prepareRename", params, callback)

    async def text_document_prepare_rename_async(
        self,
        params: types.PrepareRenameParams,
    ) -> Union[types.Range, types.PrepareRenameResult_Type1, types.PrepareRenameResult_Type2, None]:
        """Make a :lsp:`textDocument/prepareRename` request.

        A request to test and perform the setup necessary for a rename.

        @since 3.16 - support for default behavior
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/prepareRename", params)

    def text_document_prepare_type_hierarchy(
        self,
        params: types.TypeHierarchyPrepareParams,
        callback: Optional[Callable[[Optional[List[types.TypeHierarchyItem]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/prepareTypeHierarchy` request.

        A request to result a `TypeHierarchyItem` in a document at a given position.
        Can be used as an input to a subtypes or supertypes type hierarchy.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/prepareTypeHierarchy", params, callback)

    async def text_document_prepare_type_hierarchy_async(
        self,
        params: types.TypeHierarchyPrepareParams,
    ) -> Optional[List[types.TypeHierarchyItem]]:
        """Make a :lsp:`textDocument/prepareTypeHierarchy` request.

        A request to result a `TypeHierarchyItem` in a document at a given position.
        Can be used as an input to a subtypes or supertypes type hierarchy.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/prepareTypeHierarchy", params)

    def text_document_ranges_formatting(
        self,
        params: types.DocumentRangesFormattingParams,
        callback: Optional[Callable[[Optional[List[types.TextEdit]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/rangesFormatting` request.

        A request to format ranges in a document.

        @since 3.18.0
        @proposed
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/rangesFormatting", params, callback)

    async def text_document_ranges_formatting_async(
        self,
        params: types.DocumentRangesFormattingParams,
    ) -> Optional[List[types.TextEdit]]:
        """Make a :lsp:`textDocument/rangesFormatting` request.

        A request to format ranges in a document.

        @since 3.18.0
        @proposed
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/rangesFormatting", params)

    def text_document_range_formatting(
        self,
        params: types.DocumentRangeFormattingParams,
        callback: Optional[Callable[[Optional[List[types.TextEdit]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/rangeFormatting` request.

        A request to format a range in a document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/rangeFormatting", params, callback)

    async def text_document_range_formatting_async(
        self,
        params: types.DocumentRangeFormattingParams,
    ) -> Optional[List[types.TextEdit]]:
        """Make a :lsp:`textDocument/rangeFormatting` request.

        A request to format a range in a document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/rangeFormatting", params)

    def text_document_references(
        self,
        params: types.ReferenceParams,
        callback: Optional[Callable[[Optional[List[types.Location]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/references` request.

        A request to resolve project-wide references for the symbol denoted
        by the given text document position. The request's parameter is of
        type {@link ReferenceParams} the response is of type
        {@link Location Location[]} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/references", params, callback)

    async def text_document_references_async(
        self,
        params: types.ReferenceParams,
    ) -> Optional[List[types.Location]]:
        """Make a :lsp:`textDocument/references` request.

        A request to resolve project-wide references for the symbol denoted
        by the given text document position. The request's parameter is of
        type {@link ReferenceParams} the response is of type
        {@link Location Location[]} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/references", params)

    def text_document_rename(
        self,
        params: types.RenameParams,
        callback: Optional[Callable[[Optional[types.WorkspaceEdit]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/rename` request.

        A request to rename a symbol.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/rename", params, callback)

    async def text_document_rename_async(
        self,
        params: types.RenameParams,
    ) -> Optional[types.WorkspaceEdit]:
        """Make a :lsp:`textDocument/rename` request.

        A request to rename a symbol.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/rename", params)

    def text_document_selection_range(
        self,
        params: types.SelectionRangeParams,
        callback: Optional[Callable[[Optional[List[types.SelectionRange]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/selectionRange` request.

        A request to provide selection ranges in a document. The request's
        parameter is of type {@link SelectionRangeParams}, the
        response is of type {@link SelectionRange SelectionRange[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/selectionRange", params, callback)

    async def text_document_selection_range_async(
        self,
        params: types.SelectionRangeParams,
    ) -> Optional[List[types.SelectionRange]]:
        """Make a :lsp:`textDocument/selectionRange` request.

        A request to provide selection ranges in a document. The request's
        parameter is of type {@link SelectionRangeParams}, the
        response is of type {@link SelectionRange SelectionRange[]} or a Thenable
        that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/selectionRange", params)

    def text_document_semantic_tokens_full(
        self,
        params: types.SemanticTokensParams,
        callback: Optional[Callable[[Optional[types.SemanticTokens]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/semanticTokens/full` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/semanticTokens/full", params, callback)

    async def text_document_semantic_tokens_full_async(
        self,
        params: types.SemanticTokensParams,
    ) -> Optional[types.SemanticTokens]:
        """Make a :lsp:`textDocument/semanticTokens/full` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/semanticTokens/full", params)

    def text_document_semantic_tokens_full_delta(
        self,
        params: types.SemanticTokensDeltaParams,
        callback: Optional[Callable[[Union[types.SemanticTokens, types.SemanticTokensDelta, None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/semanticTokens/full/delta` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/semanticTokens/full/delta", params, callback)

    async def text_document_semantic_tokens_full_delta_async(
        self,
        params: types.SemanticTokensDeltaParams,
    ) -> Union[types.SemanticTokens, types.SemanticTokensDelta, None]:
        """Make a :lsp:`textDocument/semanticTokens/full/delta` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/semanticTokens/full/delta", params)

    def text_document_semantic_tokens_range(
        self,
        params: types.SemanticTokensRangeParams,
        callback: Optional[Callable[[Optional[types.SemanticTokens]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/semanticTokens/range` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/semanticTokens/range", params, callback)

    async def text_document_semantic_tokens_range_async(
        self,
        params: types.SemanticTokensRangeParams,
    ) -> Optional[types.SemanticTokens]:
        """Make a :lsp:`textDocument/semanticTokens/range` request.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/semanticTokens/range", params)

    def text_document_signature_help(
        self,
        params: types.SignatureHelpParams,
        callback: Optional[Callable[[Optional[types.SignatureHelp]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/signatureHelp` request.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/signatureHelp", params, callback)

    async def text_document_signature_help_async(
        self,
        params: types.SignatureHelpParams,
    ) -> Optional[types.SignatureHelp]:
        """Make a :lsp:`textDocument/signatureHelp` request.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/signatureHelp", params)

    def text_document_type_definition(
        self,
        params: types.TypeDefinitionParams,
        callback: Optional[Callable[[Union[types.Location, List[types.Location], List[types.LocationLink], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/typeDefinition` request.

        A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Definition} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/typeDefinition", params, callback)

    async def text_document_type_definition_async(
        self,
        params: types.TypeDefinitionParams,
    ) -> Union[types.Location, List[types.Location], List[types.LocationLink], None]:
        """Make a :lsp:`textDocument/typeDefinition` request.

        A request to resolve the type definition locations of a symbol at a given text
        document position. The request's parameter is of type {@link TextDocumentPositionParams}
        the response is of type {@link Definition} or a Thenable that resolves to such.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/typeDefinition", params)

    def text_document_will_save_wait_until(
        self,
        params: types.WillSaveTextDocumentParams,
        callback: Optional[Callable[[Optional[List[types.TextEdit]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`textDocument/willSaveWaitUntil` request.

        A document will save request is sent from the client to the server before
        the document is actually saved. The request can return an array of TextEdits
        which will be applied to the text document before it is saved. Please note that
        clients might drop results if computing the text edits took too long or if a
        server constantly fails on this request. This is done to keep the save fast and
        reliable.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("textDocument/willSaveWaitUntil", params, callback)

    async def text_document_will_save_wait_until_async(
        self,
        params: types.WillSaveTextDocumentParams,
    ) -> Optional[List[types.TextEdit]]:
        """Make a :lsp:`textDocument/willSaveWaitUntil` request.

        A document will save request is sent from the client to the server before
        the document is actually saved. The request can return an array of TextEdits
        which will be applied to the text document before it is saved. Please note that
        clients might drop results if computing the text edits took too long or if a
        server constantly fails on this request. This is done to keep the save fast and
        reliable.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("textDocument/willSaveWaitUntil", params)

    def type_hierarchy_subtypes(
        self,
        params: types.TypeHierarchySubtypesParams,
        callback: Optional[Callable[[Optional[List[types.TypeHierarchyItem]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`typeHierarchy/subtypes` request.

        A request to resolve the subtypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("typeHierarchy/subtypes", params, callback)

    async def type_hierarchy_subtypes_async(
        self,
        params: types.TypeHierarchySubtypesParams,
    ) -> Optional[List[types.TypeHierarchyItem]]:
        """Make a :lsp:`typeHierarchy/subtypes` request.

        A request to resolve the subtypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("typeHierarchy/subtypes", params)

    def type_hierarchy_supertypes(
        self,
        params: types.TypeHierarchySupertypesParams,
        callback: Optional[Callable[[Optional[List[types.TypeHierarchyItem]]], None]] = None,
    ) -> Future:
        """Make a :lsp:`typeHierarchy/supertypes` request.

        A request to resolve the supertypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("typeHierarchy/supertypes", params, callback)

    async def type_hierarchy_supertypes_async(
        self,
        params: types.TypeHierarchySupertypesParams,
    ) -> Optional[List[types.TypeHierarchyItem]]:
        """Make a :lsp:`typeHierarchy/supertypes` request.

        A request to resolve the supertypes for a given `TypeHierarchyItem`.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("typeHierarchy/supertypes", params)

    def workspace_diagnostic(
        self,
        params: types.WorkspaceDiagnosticParams,
        callback: Optional[Callable[[types.WorkspaceDiagnosticReport], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/diagnostic` request.

        The workspace diagnostic request definition.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/diagnostic", params, callback)

    async def workspace_diagnostic_async(
        self,
        params: types.WorkspaceDiagnosticParams,
    ) -> types.WorkspaceDiagnosticReport:
        """Make a :lsp:`workspace/diagnostic` request.

        The workspace diagnostic request definition.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/diagnostic", params)

    def workspace_execute_command(
        self,
        params: types.ExecuteCommandParams,
        callback: Optional[Callable[[Optional[Any]], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/executeCommand` request.

        A request send from the client to the server to execute a command. The request might return
        a workspace edit which the client will apply to the workspace.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/executeCommand", params, callback)

    async def workspace_execute_command_async(
        self,
        params: types.ExecuteCommandParams,
    ) -> Optional[Any]:
        """Make a :lsp:`workspace/executeCommand` request.

        A request send from the client to the server to execute a command. The request might return
        a workspace edit which the client will apply to the workspace.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/executeCommand", params)

    def workspace_symbol(
        self,
        params: types.WorkspaceSymbolParams,
        callback: Optional[Callable[[Union[List[types.SymbolInformation], List[types.WorkspaceSymbol], None]], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/symbol` request.

        A request to list project-wide symbols matching the query string given
        by the {@link WorkspaceSymbolParams}. The response is
        of type {@link SymbolInformation SymbolInformation[]} or a Thenable that
        resolves to such.

        @since 3.17.0 - support for WorkspaceSymbol in the returned data. Clients
         need to advertise support for WorkspaceSymbols via the client capability
         `workspace.symbol.resolveSupport`.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/symbol", params, callback)

    async def workspace_symbol_async(
        self,
        params: types.WorkspaceSymbolParams,
    ) -> Union[List[types.SymbolInformation], List[types.WorkspaceSymbol], None]:
        """Make a :lsp:`workspace/symbol` request.

        A request to list project-wide symbols matching the query string given
        by the {@link WorkspaceSymbolParams}. The response is
        of type {@link SymbolInformation SymbolInformation[]} or a Thenable that
        resolves to such.

        @since 3.17.0 - support for WorkspaceSymbol in the returned data. Clients
         need to advertise support for WorkspaceSymbols via the client capability
         `workspace.symbol.resolveSupport`.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/symbol", params)

    def workspace_symbol_resolve(
        self,
        params: types.WorkspaceSymbol,
        callback: Optional[Callable[[types.WorkspaceSymbol], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspaceSymbol/resolve` request.

        A request to resolve the range inside the workspace
        symbol's location.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspaceSymbol/resolve", params, callback)

    async def workspace_symbol_resolve_async(
        self,
        params: types.WorkspaceSymbol,
    ) -> types.WorkspaceSymbol:
        """Make a :lsp:`workspaceSymbol/resolve` request.

        A request to resolve the range inside the workspace
        symbol's location.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspaceSymbol/resolve", params)

    def workspace_will_create_files(
        self,
        params: types.CreateFilesParams,
        callback: Optional[Callable[[Optional[types.WorkspaceEdit]], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/willCreateFiles` request.

        The will create files request is sent from the client to the server before files are actually
        created as long as the creation is triggered from within the client.

        The request can return a `WorkspaceEdit` which will be applied to workspace before the
        files are created. Hence the `WorkspaceEdit` can not manipulate the content of the file
        to be created.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/willCreateFiles", params, callback)

    async def workspace_will_create_files_async(
        self,
        params: types.CreateFilesParams,
    ) -> Optional[types.WorkspaceEdit]:
        """Make a :lsp:`workspace/willCreateFiles` request.

        The will create files request is sent from the client to the server before files are actually
        created as long as the creation is triggered from within the client.

        The request can return a `WorkspaceEdit` which will be applied to workspace before the
        files are created. Hence the `WorkspaceEdit` can not manipulate the content of the file
        to be created.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/willCreateFiles", params)

    def workspace_will_delete_files(
        self,
        params: types.DeleteFilesParams,
        callback: Optional[Callable[[Optional[types.WorkspaceEdit]], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/willDeleteFiles` request.

        The did delete files notification is sent from the client to the server when
        files were deleted from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/willDeleteFiles", params, callback)

    async def workspace_will_delete_files_async(
        self,
        params: types.DeleteFilesParams,
    ) -> Optional[types.WorkspaceEdit]:
        """Make a :lsp:`workspace/willDeleteFiles` request.

        The did delete files notification is sent from the client to the server when
        files were deleted from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/willDeleteFiles", params)

    def workspace_will_rename_files(
        self,
        params: types.RenameFilesParams,
        callback: Optional[Callable[[Optional[types.WorkspaceEdit]], None]] = None,
    ) -> Future:
        """Make a :lsp:`workspace/willRenameFiles` request.

        The will rename files request is sent from the client to the server before files are actually
        renamed as long as the rename is triggered from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return self.protocol.send_request("workspace/willRenameFiles", params, callback)

    async def workspace_will_rename_files_async(
        self,
        params: types.RenameFilesParams,
    ) -> Optional[types.WorkspaceEdit]:
        """Make a :lsp:`workspace/willRenameFiles` request.

        The will rename files request is sent from the client to the server before files are actually
        renamed as long as the rename is triggered from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        return await self.protocol.send_request_async("workspace/willRenameFiles", params)

    def cancel_request(self, params: types.CancelParams) -> None:
        """Send a :lsp:`$/cancelRequest` notification.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("$/cancelRequest", params)

    def exit(self, params: None) -> None:
        """Send a :lsp:`exit` notification.

        The exit event is sent from the client to the server to
        ask the server to exit its process.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("exit", params)

    def initialized(self, params: types.InitializedParams) -> None:
        """Send a :lsp:`initialized` notification.

        The initialized notification is sent from the client to the
        server after the client is fully initialized and the server
        is allowed to send requests from the server to the client.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("initialized", params)

    def notebook_document_did_change(self, params: types.DidChangeNotebookDocumentParams) -> None:
        """Send a :lsp:`notebookDocument/didChange` notification.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("notebookDocument/didChange", params)

    def notebook_document_did_close(self, params: types.DidCloseNotebookDocumentParams) -> None:
        """Send a :lsp:`notebookDocument/didClose` notification.

        A notification sent when a notebook closes.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("notebookDocument/didClose", params)

    def notebook_document_did_open(self, params: types.DidOpenNotebookDocumentParams) -> None:
        """Send a :lsp:`notebookDocument/didOpen` notification.

        A notification sent when a notebook opens.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("notebookDocument/didOpen", params)

    def notebook_document_did_save(self, params: types.DidSaveNotebookDocumentParams) -> None:
        """Send a :lsp:`notebookDocument/didSave` notification.

        A notification sent when a notebook document is saved.

        @since 3.17.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("notebookDocument/didSave", params)

    def progress(self, params: types.ProgressParams) -> None:
        """Send a :lsp:`$/progress` notification.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("$/progress", params)

    def set_trace(self, params: types.SetTraceParams) -> None:
        """Send a :lsp:`$/setTrace` notification.


        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("$/setTrace", params)

    def text_document_did_change(self, params: types.DidChangeTextDocumentParams) -> None:
        """Send a :lsp:`textDocument/didChange` notification.

        The document change notification is sent from the client to the server to signal
        changes to a text document.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("textDocument/didChange", params)

    def text_document_did_close(self, params: types.DidCloseTextDocumentParams) -> None:
        """Send a :lsp:`textDocument/didClose` notification.

        The document close notification is sent from the client to the server when
        the document got closed in the client. The document's truth now exists where
        the document's uri points to (e.g. if the document's uri is a file uri the
        truth now exists on disk). As with the open notification the close notification
        is about managing the document's content. Receiving a close notification
        doesn't mean that the document was open in an editor before. A close
        notification requires a previous open notification to be sent.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("textDocument/didClose", params)

    def text_document_did_open(self, params: types.DidOpenTextDocumentParams) -> None:
        """Send a :lsp:`textDocument/didOpen` notification.

        The document open notification is sent from the client to the server to signal
        newly opened text documents. The document's truth is now managed by the client
        and the server must not try to read the document's truth using the document's
        uri. Open in this sense means it is managed by the client. It doesn't necessarily
        mean that its content is presented in an editor. An open notification must not
        be sent more than once without a corresponding close notification send before.
        This means open and close notification must be balanced and the max open count
        is one.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("textDocument/didOpen", params)

    def text_document_did_save(self, params: types.DidSaveTextDocumentParams) -> None:
        """Send a :lsp:`textDocument/didSave` notification.

        The document save notification is sent from the client to the server when
        the document got saved in the client.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("textDocument/didSave", params)

    def text_document_will_save(self, params: types.WillSaveTextDocumentParams) -> None:
        """Send a :lsp:`textDocument/willSave` notification.

        A document will save notification is sent from the client to the server before
        the document is actually saved.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("textDocument/willSave", params)

    def window_work_done_progress_cancel(self, params: types.WorkDoneProgressCancelParams) -> None:
        """Send a :lsp:`window/workDoneProgress/cancel` notification.

        The `window/workDoneProgress/cancel` notification is sent from  the client to the server to cancel a progress
        initiated on the server side.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("window/workDoneProgress/cancel", params)

    def workspace_did_change_configuration(self, params: types.DidChangeConfigurationParams) -> None:
        """Send a :lsp:`workspace/didChangeConfiguration` notification.

        The configuration change notification is sent from the client to the server
        when the client's configuration has changed. The notification contains
        the changed configuration as defined by the language client.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didChangeConfiguration", params)

    def workspace_did_change_watched_files(self, params: types.DidChangeWatchedFilesParams) -> None:
        """Send a :lsp:`workspace/didChangeWatchedFiles` notification.

        The watched files notification is sent from the client to the server when
        the client detects changes to file watched by the language client.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didChangeWatchedFiles", params)

    def workspace_did_change_workspace_folders(self, params: types.DidChangeWorkspaceFoldersParams) -> None:
        """Send a :lsp:`workspace/didChangeWorkspaceFolders` notification.

        The `workspace/didChangeWorkspaceFolders` notification is sent from the client to the server when the workspace
        folder configuration changes.
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didChangeWorkspaceFolders", params)

    def workspace_did_create_files(self, params: types.CreateFilesParams) -> None:
        """Send a :lsp:`workspace/didCreateFiles` notification.

        The did create files notification is sent from the client to the server when
        files were created from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didCreateFiles", params)

    def workspace_did_delete_files(self, params: types.DeleteFilesParams) -> None:
        """Send a :lsp:`workspace/didDeleteFiles` notification.

        The will delete files request is sent from the client to the server before files are actually
        deleted as long as the deletion is triggered from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didDeleteFiles", params)

    def workspace_did_rename_files(self, params: types.RenameFilesParams) -> None:
        """Send a :lsp:`workspace/didRenameFiles` notification.

        The did rename files notification is sent from the client to the server when
        files were renamed from within the client.

        @since 3.16.0
        """
        if self.stopped:
            raise RuntimeError("Client has been stopped.")

        self.protocol.notify("workspace/didRenameFiles", params)



import logging
from typing import Final

def byte_ord(c: int | str) -> int: ...
def byte_chr(c: int) -> bytes: ...
def byte_mask(c: int, mask: int) -> bytes: ...

MSG_DISCONNECT: Final = 1
MSG_IGNORE: Final = 2
MSG_UNIMPLEMENTED: Final = 3
MSG_DEBUG: Final = 4
MSG_SERVICE_REQUEST: Final = 5
MSG_SERVICE_ACCEPT: Final = 6
MSG_EXT_INFO: Final = 7
MSG_KEXINIT: Final = 20
MSG_NEWKEYS: Final = 21
MSG_USERAUTH_REQUEST: Final = 50
MSG_USERAUTH_FAILURE: Final = 51
MSG_USERAUTH_SUCCESS: Final = 52
MSG_USERAUTH_BANNER: Final = 53
MSG_USERAUTH_PK_OK: Final = 60
MSG_USERAUTH_INFO_REQUEST: Final = 60
MSG_USERAUTH_INFO_RESPONSE: Final = 61
MSG_USERAUTH_GSSAPI_RESPONSE: Final = 60
MSG_USERAUTH_GSSAPI_TOKEN: Final = 61
MSG_USERAUTH_GSSAPI_EXCHANGE_COMPLETE: Final = 63
MSG_USERAUTH_GSSAPI_ERROR: Final = 64
MSG_USERAUTH_GSSAPI_ERRTOK: Final = 65
MSG_USERAUTH_GSSAPI_MIC: Final = 66
HIGHEST_USERAUTH_MESSAGE_ID: Final = 79
MSG_GLOBAL_REQUEST: Final = 80
MSG_REQUEST_SUCCESS: Final = 81
MSG_REQUEST_FAILURE: Final = 82
MSG_CHANNEL_OPEN: Final = 90
MSG_CHANNEL_OPEN_SUCCESS: Final = 91
MSG_CHANNEL_OPEN_FAILURE: Final = 92
MSG_CHANNEL_WINDOW_ADJUST: Final = 93
MSG_CHANNEL_DATA: Final = 94
MSG_CHANNEL_EXTENDED_DATA: Final = 95
MSG_CHANNEL_EOF: Final = 96
MSG_CHANNEL_CLOSE: Final = 97
MSG_CHANNEL_REQUEST: Final = 98
MSG_CHANNEL_SUCCESS: Final = 99
MSG_CHANNEL_FAILURE: Final = 100

cMSG_DISCONNECT: Final[bytes]
cMSG_IGNORE: Final[bytes]
cMSG_UNIMPLEMENTED: Final[bytes]
cMSG_DEBUG: Final[bytes]
cMSG_SERVICE_REQUEST: Final[bytes]
cMSG_SERVICE_ACCEPT: Final[bytes]
cMSG_EXT_INFO: Final[bytes]
cMSG_KEXINIT: Final[bytes]
cMSG_NEWKEYS: Final[bytes]
cMSG_USERAUTH_REQUEST: Final[bytes]
cMSG_USERAUTH_FAILURE: Final[bytes]
cMSG_USERAUTH_SUCCESS: Final[bytes]
cMSG_USERAUTH_BANNER: Final[bytes]
cMSG_USERAUTH_PK_OK: Final[bytes]
cMSG_USERAUTH_INFO_REQUEST: Final[bytes]
cMSG_USERAUTH_INFO_RESPONSE: Final[bytes]
cMSG_USERAUTH_GSSAPI_RESPONSE: Final[bytes]
cMSG_USERAUTH_GSSAPI_TOKEN: Final[bytes]
cMSG_USERAUTH_GSSAPI_EXCHANGE_COMPLETE: Final[bytes]
cMSG_USERAUTH_GSSAPI_ERROR: Final[bytes]
cMSG_USERAUTH_GSSAPI_ERRTOK: Final[bytes]
cMSG_USERAUTH_GSSAPI_MIC: Final[bytes]
cMSG_GLOBAL_REQUEST: Final[bytes]
cMSG_REQUEST_SUCCESS: Final[bytes]
cMSG_REQUEST_FAILURE: Final[bytes]
cMSG_CHANNEL_OPEN: Final[bytes]
cMSG_CHANNEL_OPEN_SUCCESS: Final[bytes]
cMSG_CHANNEL_OPEN_FAILURE: Final[bytes]
cMSG_CHANNEL_WINDOW_ADJUST: Final[bytes]
cMSG_CHANNEL_DATA: Final[bytes]
cMSG_CHANNEL_EXTENDED_DATA: Final[bytes]
cMSG_CHANNEL_EOF: Final[bytes]
cMSG_CHANNEL_CLOSE: Final[bytes]
cMSG_CHANNEL_REQUEST: Final[bytes]
cMSG_CHANNEL_SUCCESS: Final[bytes]
cMSG_CHANNEL_FAILURE: Final[bytes]

MSG_NAMES: dict[int, str]

AUTH_SUCCESSFUL: Final = 0
AUTH_PARTIALLY_SUCCESSFUL: Final = 1
AUTH_FAILED: Final = 2

OPEN_SUCCEEDED: Final = 0
OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED: Final = 1
OPEN_FAILED_CONNECT_FAILED: Final = 2
OPEN_FAILED_UNKNOWN_CHANNEL_TYPE: Final = 3
OPEN_FAILED_RESOURCE_SHORTAGE: Final = 4

CONNECTION_FAILED_CODE: dict[int, str]

DISCONNECT_SERVICE_NOT_AVAILABLE: Final = 7
DISCONNECT_AUTH_CANCELLED_BY_USER: Final = 13
DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE: Final = 14

zero_byte: bytes
one_byte: bytes
four_byte: bytes
max_byte: bytes
cr_byte: bytes
linefeed_byte: bytes
crlf: bytes
cr_byte_value: int
linefeed_byte_value: int
xffffffff: int
x80000000: int
o666: int
o660: int
o644: int
o600: int
o777: int
o700: int
o70: int

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

io_sleep: float

DEFAULT_WINDOW_SIZE: Final[int]
DEFAULT_MAX_PACKET_SIZE: Final[int]

MIN_WINDOW_SIZE: Final[int]
MIN_PACKET_SIZE: Final[int]
MAX_WINDOW_SIZE: Final[int]



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["communicability", "communicability_exp"]

@_dispatchable
def communicability(G: Graph[_Node]) -> dict[dict[Incomplete, Incomplete], dict[Incomplete, float]]: ...
@_dispatchable
def communicability_exp(G: Graph[_Node]) -> dict[dict[Incomplete, Incomplete], dict[Incomplete, float]]: ...



from _typeshed.wsgi import WSGIApplication
from collections.abc import Mapping, Sequence
from logging import Logger
from typing import Final, NamedTuple, TypeVar

_T = TypeVar("_T")

PROXY_HEADERS: Final[frozenset[str]]

class Forwarded(NamedTuple):
    by: str
    for_: str
    host: str
    proto: str

class MalformedProxyHeader(Exception):
    header: str
    reason: str
    value: str
    def __init__(self, header: str, reason: str, value: str) -> None: ...

def proxy_headers_middleware(
    app: WSGIApplication,
    trusted_proxy: str | None = None,
    trusted_proxy_count: int = 1,
    trusted_proxy_headers: set[str] | None = None,
    clear_untrusted: bool = True,
    log_untrusted: bool = False,
    logger: Logger = ...,
) -> WSGIApplication: ...
def parse_proxy_headers(
    environ: Mapping[str, str], trusted_proxy_count: int, trusted_proxy_headers: set[str], logger: Logger = ...
) -> set[str]: ...
def strip_brackets(addr: Sequence[_T]) -> Sequence[_T]: ...
def clear_untrusted_headers(
    environ: Mapping[str, str], untrusted_headers: Sequence[str], log_warning: bool = False, logger: Logger = ...
) -> None: ...



from _typeshed import Incomplete
from collections.abc import Callable, Container, Iterator
from pathlib import Path
from typing import Any, Literal
from typing_extensions import Self, TypeAlias, deprecated

import numpy as np
import numpy.typing as npt
import tensorflow as tf
from tensorflow import Variable
from tensorflow._aliases import ContainerGeneric, ShapeLike, TensorCompatible
from tensorflow.keras.layers import Layer, _InputT_contra, _OutputT_co
from tensorflow.keras.optimizers import Optimizer

_Loss: TypeAlias = str | tf.keras.losses.Loss | Callable[[TensorCompatible, TensorCompatible], tf.Tensor]
_Metric: TypeAlias = str | tf.keras.metrics.Metric | Callable[[TensorCompatible, TensorCompatible], tf.Tensor] | None

# Missing keras.src.backend.tensorflow.trainer.TensorFlowTrainer as a base class, which is not exposed by tensorflow
class Model(Layer[_InputT_contra, _OutputT_co]):
    _train_counter: tf.Variable
    _test_counter: tf.Variable
    optimizer: Optimizer | None
    # This is actually TensorFlowTrainer.loss
    @deprecated("Instead, use `model.compute_loss(x, y, y_pred, sample_weight)`.")
    def loss(self, y: TensorCompatible | None, y_pred: TensorCompatible | None, sample_weight=None) -> tf.Tensor | None: ...
    stop_training: bool

    def __new__(cls, *args: Any, **kwargs: Any) -> Model[_InputT_contra, _OutputT_co]: ...
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def __setattr__(self, name: str, value: Any) -> None: ...
    def __reduce__(self): ...
    def build(self, input_shape: ShapeLike) -> None: ...
    def __call__(
        self, inputs: _InputT_contra, *, training: bool = False, mask: TensorCompatible | None = None
    ) -> _OutputT_co: ...
    def call(self, inputs: _InputT_contra, training: bool | None = None, mask: TensorCompatible | None = None) -> _OutputT_co: ...
    # Ideally loss/metrics/output would share the same structure but higher kinded types are not supported.
    def compile(
        self,
        optimizer: Optimizer | str = "rmsprop",
        loss: ContainerGeneric[_Loss] | None = None,
        loss_weights: ContainerGeneric[float] | None = None,
        metrics: ContainerGeneric[_Metric] | None = None,
        weighted_metrics: ContainerGeneric[_Metric] | None = None,
        run_eagerly: bool = False,
        steps_per_execution: int | Literal["auto"] = 1,
        jit_compile: bool | Literal["auto"] = "auto",
        auto_scale_loss: bool | None = True,
    ) -> None: ...
    @property
    def metrics(self) -> list[Incomplete]: ...
    @property
    def metrics_names(self) -> list[str]: ...
    @property
    def distribute_strategy(self) -> tf.distribute.Strategy: ...
    @property
    def run_eagerly(self) -> bool: ...
    @property
    def jit_compile(self) -> bool: ...
    @property
    def distribute_reduction_method(self) -> Incomplete | Literal["auto"]: ...
    def train_step(self, data: TensorCompatible): ...
    def compute_loss(
        self,
        x: TensorCompatible | None = None,
        y: TensorCompatible | None = None,
        y_pred: TensorCompatible | None = None,
        sample_weight=None,
        training: bool = True,
    ) -> tf.Tensor | None: ...
    def compute_metrics(
        self, x: TensorCompatible, y: TensorCompatible, y_pred: TensorCompatible, sample_weight=None
    ) -> dict[str, float]: ...
    def get_metrics_result(self) -> dict[str, float]: ...
    def make_train_function(self, force: bool = False) -> Callable[[tf.data.Iterator[Incomplete]], dict[str, float]]: ...
    def fit(
        self,
        x: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        y: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        batch_size: int | None = None,
        epochs: int = 1,
        verbose: Literal["auto", 0, 1, 2] = "auto",
        callbacks: list[tf.keras.callbacks.Callback] | None = None,
        validation_split: float = 0.0,
        validation_data: TensorCompatible | tf.data.Dataset[Any] | None = None,
        shuffle: bool = True,
        class_weight: dict[int, float] | None = None,
        sample_weight: npt.NDArray[np.float64] | None = None,
        initial_epoch: int = 0,
        steps_per_epoch: int | None = None,
        validation_steps: int | None = None,
        validation_batch_size: int | None = None,
        validation_freq: int | Container[int] = 1,
    ) -> tf.keras.callbacks.History: ...
    def test_step(self, data: TensorCompatible) -> dict[str, float]: ...
    def make_test_function(self, force: bool = False) -> Callable[[tf.data.Iterator[Incomplete]], dict[str, float]]: ...
    def evaluate(
        self,
        x: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        y: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        batch_size: int | None = None,
        verbose: Literal["auto", 0, 1, 2] = "auto",
        sample_weight: npt.NDArray[np.float64] | None = None,
        steps: int | None = None,
        callbacks: list[tf.keras.callbacks.Callback] | None = None,
        return_dict: bool = False,
        **kwargs: Any,
    ) -> float | list[float]: ...
    def predict_step(self, data: _InputT_contra) -> _OutputT_co: ...
    def make_predict_function(self, force: bool = False) -> Callable[[tf.data.Iterator[Incomplete]], _OutputT_co]: ...
    def predict(
        self,
        x: TensorCompatible | tf.data.Dataset[Incomplete],
        batch_size: int | None = None,
        verbose: Literal["auto", 0, 1, 2] = "auto",
        steps: int | None = None,
        callbacks: list[tf.keras.callbacks.Callback] | None = None,
    ) -> _OutputT_co: ...
    def reset_metrics(self) -> None: ...
    def train_on_batch(
        self,
        x: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete],
        y: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        sample_weight: npt.NDArray[np.float64] | None = None,
        class_weight: dict[int, float] | None = None,
        return_dict: bool = False,
    ) -> float | list[float]: ...
    def test_on_batch(
        self,
        x: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete],
        y: TensorCompatible | dict[str, TensorCompatible] | tf.data.Dataset[Incomplete] | None = None,
        sample_weight: npt.NDArray[np.float64] | None = None,
        return_dict: bool = False,
    ) -> float | list[float]: ...
    def predict_on_batch(self, x: Iterator[_InputT_contra]) -> npt.NDArray[Incomplete]: ...
    @property
    def trainable_weights(self) -> list[Variable]: ...
    @property
    def non_trainable_weights(self) -> list[Variable]: ...
    def get_weights(self): ...
    def save(self, filepath: str | Path, overwrite: bool = True, zipped: bool | None = None) -> None: ...
    def save_weights(self, filepath: str | Path, overwrite: bool = True) -> None: ...
    # kwargs are from keras.saving.saving_api.load_weights
    def load_weights(self, filepath: str | Path, skip_mismatch: bool = False, *, by_name: bool = False) -> None: ...
    def get_config(self) -> dict[str, Any]: ...
    @classmethod
    def from_config(cls, config: dict[str, Any], custom_objects=None) -> Self: ...
    def to_json(self, **kwargs: Any) -> str: ...
    @property
    def weights(self) -> list[Variable]: ...
    def summary(
        self,
        line_length: None | int = None,
        positions: None | list[float] = None,
        print_fn: None | Callable[[str], None] = None,
        expand_nested: bool = False,
        show_trainable: bool = False,
        layer_range: None | list[str] | tuple[str, str] = None,
    ) -> None: ...
    @property
    def layers(self) -> list[Layer[Incomplete, Incomplete]]: ...
    def get_layer(self, name: str | None = None, index: int | None = None) -> Layer[Incomplete, Incomplete]: ...
    def get_compile_config(self) -> dict[str, Any]: ...
    def compile_from_config(self, config: dict[str, Any]) -> Self: ...
    def export(self, filepath: str | Path, format: str = "tf_saved_model", verbose: bool = True) -> None: ...

def __getattr__(name: str): ...  # incomplete module



from auth0.rest import RestClientOptions

from .actions import Actions
from .attack_protection import AttackProtection
from .blacklists import Blacklists
from .branding import Branding
from .client_credentials import ClientCredentials
from .client_grants import ClientGrants
from .clients import Clients
from .connections import Connections
from .custom_domains import CustomDomains
from .device_credentials import DeviceCredentials
from .email_templates import EmailTemplates
from .emails import Emails
from .grants import Grants
from .guardian import Guardian
from .hooks import Hooks
from .jobs import Jobs
from .log_streams import LogStreams
from .logs import Logs
from .organizations import Organizations
from .prompts import Prompts
from .resource_servers import ResourceServers
from .roles import Roles
from .rules import Rules
from .rules_configs import RulesConfigs
from .self_service_profiles import SelfServiceProfiles
from .stats import Stats
from .tenants import Tenants
from .tickets import Tickets
from .user_blocks import UserBlocks
from .users import Users
from .users_by_email import UsersByEmail

class Auth0:
    actions: Actions
    attack_protection: AttackProtection
    blacklists: Blacklists
    branding: Branding
    client_credentials: ClientCredentials
    client_grants: ClientGrants
    clients: Clients
    connections: Connections
    custom_domains: CustomDomains
    device_credentials: DeviceCredentials
    email_templates: EmailTemplates
    emails: Emails
    grants: Grants
    guardian: Guardian
    hooks: Hooks
    jobs: Jobs
    log_streams: LogStreams
    logs: Logs
    organizations: Organizations
    prompts: Prompts
    resource_servers: ResourceServers
    roles: Roles
    rules_configs: RulesConfigs
    rules: Rules
    self_service_profiles: SelfServiceProfiles
    stats: Stats
    tenants: Tenants
    tickets: Tickets
    user_blocks: UserBlocks
    users_by_email: UsersByEmail
    users: Users
    def __init__(self, domain: str, token: str, rest_options: RestClientOptions | None = None) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Protocol buffer representing the shape of tensors."""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class TensorShapeProto(google.protobuf.message.Message):
    """Dimensions of a tensor."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Dim(google.protobuf.message.Message):
        """One dimension of the tensor."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SIZE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        size: builtins.int
        """Size of the tensor in that dimension.
        This value must be >= -1, but values of -1 are reserved for "unknown"
        shapes (values of -1 mean "unknown" dimension).  Certain wrappers
        that work with TensorShapeProto may fail at runtime when deserializing
        a TensorShapeProto containing a dim value of -1.
        """
        name: builtins.str
        """Optional name of the tensor dimension."""
        def __init__(self, *, size: builtins.int | None = ..., name: builtins.str | None = ...) -> None: ...
        def ClearField(self, field_name: typing.Literal["name", b"name", "size", b"size"]) -> None: ...

    DIM_FIELD_NUMBER: builtins.int
    UNKNOWN_RANK_FIELD_NUMBER: builtins.int
    unknown_rank: builtins.bool
    """If true, the number of dimensions in the shape is unknown.

    If true, "dim.size()" must be 0.
    """
    @property
    def dim(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TensorShapeProto.Dim]:
        """Dimensions of the tensor, such as {"input", 30}, {"output", 40}
        for a 30 x 40 2D tensor.  If an entry has size -1, this
        corresponds to a dimension of unknown size. The names are
        optional.

        The order of entries in "dim" matters: It indicates the layout of the
        values in the tensor in-memory representation.

        The first entry in "dim" is the outermost dimension used to layout the
        values, the last entry is the innermost dimension.  This matches the
        in-memory layout of RowMajor Eigen tensors.

        If "dim.size()" > 0, "unknown_rank" must be false.
        """

    def __init__(
        self,
        *,
        dim: collections.abc.Iterable[global___TensorShapeProto.Dim] | None = ...,
        unknown_rank: builtins.bool | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dim", b"dim", "unknown_rank", b"unknown_rank"]) -> None: ...

global___TensorShapeProto = TensorShapeProto



from collections.abc import Callable
from typing import TypeVar
from typing_extensions import ParamSpec

_P = ParamSpec("_P")
_T = TypeVar("_T")

class RateLimitDecorator:
    def __init__(
        self, calls: int = 15, period: float = 900, clock: Callable[[], float] = ..., raise_on_limit: bool = True
    ) -> None: ...
    def __call__(self, func: Callable[_P, _T]) -> Callable[_P, _T]: ...

def sleep_and_retry(func: Callable[_P, _T]) -> Callable[_P, _T]: ...



import datetime
from typing import Any
from typing_extensions import TypeAlias

from ..core import Calendar

_HebrewDate: TypeAlias = Any | datetime.date  # from `pyluach.dates` package

class Israel(Calendar):
    def get_hebrew_independence_day(self, jewish_year: int) -> list[tuple[_HebrewDate, str]]: ...



from .base import DIRKey as DIRKey
from .cryptography_backend import (
    CryptographyAESKey as CryptographyAESKey,
    CryptographyECKey as CryptographyECKey,
    CryptographyHMACKey as CryptographyHMACKey,
    CryptographyRSAKey as CryptographyRSAKey,
)
from .ecdsa_backend import ECDSAECKey as ECDSAECKey
from .native import HMACKey as NativeHMACKey, get_random_bytes as get_random_bytes
from .rsa_backend import RSAKey as BackendRSAKey

# python-jose relies on importing from cryptography_backend
# then falling back on other imports
# these are all the potential options
AESKey: type[CryptographyAESKey] | None
HMACKey: type[CryptographyHMACKey | NativeHMACKey]
RSAKey: type[CryptographyRSAKey | BackendRSAKey] | None
ECKey: type[CryptographyECKey | ECDSAECKey]



import sys

if sys.platform == "win32":
    import enum
    from _typeshed import Incomplete
    from collections.abc import Iterable, Iterator
    from signal import Signals
    from typing import Final, Literal, NamedTuple, TypedDict, overload, type_check_only

    from psutil import _psutil_windows
    from psutil._common import (
        ENCODING as ENCODING,
        AccessDenied as AccessDenied,
        NoSuchProcess as NoSuchProcess,
        TimeoutExpired as TimeoutExpired,
        conn_tmap as conn_tmap,
        conn_to_ntuple as conn_to_ntuple,
        debug as debug,
        isfile_strict as isfile_strict,
        memoize as memoize,
        memoize_when_activated as memoize_when_activated,
        parse_environ_block as parse_environ_block,
        usage_percent as usage_percent,
    )
    from psutil._psutil_windows import (
        ABOVE_NORMAL_PRIORITY_CLASS as ABOVE_NORMAL_PRIORITY_CLASS,
        BELOW_NORMAL_PRIORITY_CLASS as BELOW_NORMAL_PRIORITY_CLASS,
        HIGH_PRIORITY_CLASS as HIGH_PRIORITY_CLASS,
        IDLE_PRIORITY_CLASS as IDLE_PRIORITY_CLASS,
        NORMAL_PRIORITY_CLASS as NORMAL_PRIORITY_CLASS,
        REALTIME_PRIORITY_CLASS as REALTIME_PRIORITY_CLASS,
    )

    from . import _common

    __extra__all__: Final[list[str]]
    CONN_DELETE_TCB: Final = "DELETE_TCB"
    ERROR_PARTIAL_COPY: Final = 299
    PYPY: Final[bool]

    class AddressFamily(enum.IntEnum):
        AF_LINK = -1

    AF_LINK: Final = AddressFamily.AF_LINK
    TCP_STATUSES: Final[dict[int, str]]

    # These noqas workaround https://github.com/astral-sh/ruff/issues/10874
    class Priority(enum.IntEnum):
        ABOVE_NORMAL_PRIORITY_CLASS = _psutil_windows.ABOVE_NORMAL_PRIORITY_CLASS  # noqa: F811
        BELOW_NORMAL_PRIORITY_CLASS = _psutil_windows.BELOW_NORMAL_PRIORITY_CLASS  # noqa: F811
        HIGH_PRIORITY_CLASS = _psutil_windows.HIGH_PRIORITY_CLASS  # noqa: F811
        IDLE_PRIORITY_CLASS = _psutil_windows.IDLE_PRIORITY_CLASS  # noqa: F811
        NORMAL_PRIORITY_CLASS = _psutil_windows.NORMAL_PRIORITY_CLASS  # noqa: F811
        REALTIME_PRIORITY_CLASS = _psutil_windows.REALTIME_PRIORITY_CLASS  # noqa: F811

    class IOPriority(enum.IntEnum):
        IOPRIO_VERYLOW = 0
        IOPRIO_LOW = 1
        IOPRIO_NORMAL = 2
        IOPRIO_HIGH = 3

    IOPRIO_VERYLOW: Final = IOPriority.IOPRIO_VERYLOW
    IOPRIO_LOW: Final = IOPriority.IOPRIO_LOW
    IOPRIO_NORMAL: Final = IOPriority.IOPRIO_NORMAL
    IOPRIO_HIGH: Final = IOPriority.IOPRIO_HIGH

    pinfo_map: Final[dict[str, int]]

    class scputimes(NamedTuple):
        user: float
        system: float
        idle: float
        interrupt: float
        dpc: float

    class svmem(NamedTuple):
        total: int
        available: int
        percent: float
        used: int
        free: int

    class pmem(NamedTuple):
        rss: int
        vms: int
        num_page_faults: int
        peak_wset: int
        wset: int
        peak_paged_pool: int
        paged_pool: int
        peak_nonpaged_pool: int
        nonpaged_pool: int
        pagefile: int
        peak_pagefile: int
        private: int

    class pfullmem(NamedTuple):
        rss: int
        vms: int
        num_page_faults: int
        peak_wset: int
        wset: int
        peak_paged_pool: int
        paged_pool: int
        peak_nonpaged_pool: int
        nonpaged_pool: int
        pagefile: int
        peak_pagefile: int
        private: int
        uss: int

    class pmmap_grouped(NamedTuple):
        path: Incomplete
        rss: Incomplete

    class pmmap_ext(NamedTuple):
        addr: Incomplete
        perms: Incomplete
        path: Incomplete
        rss: Incomplete

    class pio(NamedTuple):
        read_count: int
        write_count: int
        read_bytes: int
        write_bytes: int
        other_count: int
        other_bytes: int

    def convert_dos_path(s: str) -> str: ...
    def getpagesize() -> int: ...
    def virtual_memory() -> svmem: ...
    def swap_memory() -> _common.sswap: ...

    disk_io_counters = _psutil_windows.disk_io_counters

    def disk_usage(path: str) -> _common.sdiskusage: ...
    def disk_partitions(all: bool) -> list[_common.sdiskpart]: ...
    def cpu_times() -> scputimes: ...
    def per_cpu_times() -> list[scputimes]: ...
    def cpu_count_logical() -> int | None: ...
    def cpu_count_cores() -> int | None: ...
    def cpu_stats() -> _common.scpustats: ...
    def cpu_freq() -> list[_common.scpufreq]: ...
    def getloadavg() -> tuple[float, float, float]: ...
    @overload
    def net_connections(kind: str, _pid: Literal[-1] = -1) -> list[_common.sconn]: ...
    @overload
    def net_connections(kind: str, _pid: int = -1) -> list[_common.pconn]: ...
    def net_if_stats() -> dict[str, _common.snicstats]: ...
    def net_io_counters() -> dict[str, tuple[int, int, int, int, int, int, int, int]]: ...
    def net_if_addrs() -> list[tuple[str, int, str, str | None, None, None]]: ...
    def sensors_battery() -> _common.sbattery | None: ...
    def boot_time() -> float: ...
    def users() -> list[_common.suser]: ...
    def win_service_iter() -> Iterator[WindowsService]: ...
    def win_service_get(name: str) -> WindowsService: ...
    @type_check_only
    class _WindowsServiceAttrs(TypedDict):
        name: str
        display_name: str | None
        description: str
        binpath: str
        username: str
        start_type: str
        status: str
        pid: int | None

    class WindowsService:
        def __init__(self, name: str, display_name: str | None) -> None: ...
        def __eq__(self, other: object) -> bool: ...
        def __ne__(self, other: object) -> bool: ...
        def name(self) -> str: ...
        def display_name(self) -> str | None: ...
        def binpath(self) -> str: ...
        def username(self) -> str: ...
        def start_type(self) -> str: ...
        def pid(self) -> int: ...
        def status(self) -> str: ...
        def description(self) -> str: ...
        def as_dict(self) -> _WindowsServiceAttrs: ...

    pids = _psutil_windows.pids
    pid_exists = _psutil_windows.pid_exists
    ppid_map = _psutil_windows.ppid_map

    def is_permission_err(exc: OSError) -> bool: ...
    @overload
    def convert_oserror(exc: PermissionError, pid=None, name=None) -> AccessDenied: ...
    @overload
    def convert_oserror(exc: OSError, pid=None, name=None) -> AccessDenied | NoSuchProcess: ...
    def wrap_exceptions(fun): ...
    def retry_error_partial_copy(fun): ...

    class Process:
        __slots__ = ["_cache", "_name", "_ppid", "pid"]
        pid: int
        def __init__(self, pid: int) -> None: ...
        def oneshot_enter(self) -> None: ...
        def oneshot_exit(self) -> None: ...
        def name(self) -> str: ...
        def exe(self) -> str: ...
        def cmdline(self) -> list[str]: ...
        def environ(self) -> dict[str, str]: ...
        def ppid(self) -> int: ...
        def memory_info(self) -> pmem: ...
        def memory_full_info(self) -> pfullmem: ...
        def memory_maps(self) -> Iterator[tuple[str, str, str, int]]: ...
        def kill(self) -> None: ...
        def send_signal(self, sig: Literal[Signals.SIGTERM, Signals.CTRL_C_EVENT, Signals.CTRL_BREAK_EVENT]) -> None: ...
        def wait(self, timeout: float | None = None) -> int | None: ...
        def username(self) -> str: ...
        def create_time(self, fast_only: bool = False) -> float: ...
        def num_threads(self) -> int: ...
        def threads(self) -> list[_common.pthread]: ...
        def cpu_times(self) -> _common.pcputimes: ...
        def suspend(self) -> None: ...
        def resume(self) -> None: ...
        def cwd(self) -> str: ...
        def open_files(self) -> list[_common.popenfile]: ...
        def net_connections(self, kind: str = "inet") -> list[_common.pconn]: ...
        def nice_get(self) -> Priority: ...
        def nice_set(self, value) -> None: ...
        def ionice_get(self) -> IOPriority: ...
        def ionice_set(self, ioclass: int, value: None) -> None: ...
        def io_counters(self) -> pio: ...
        def status(self) -> Literal["stopped", "running"]: ...
        def cpu_affinity_get(self) -> list[int]: ...
        def cpu_affinity_set(self, value: Iterable[int]) -> None: ...
        def num_handles(self) -> int: ...
        def num_ctx_switches(self) -> _common.pctxsw: ...



import abc
import datetime
from collections.abc import Generator
from enum import Enum
from typing import TypedDict, overload, type_check_only
from typing_extensions import Self

@type_check_only
class _DateDict(TypedDict):
    year: int
    month: int
    day: int

class Rounding(Enum):
    PREVIOUS_DAY = 1
    NEXT_DAY = 2
    EXCEPTION = 3

class BaseDate(abc.ABC, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def jd(self) -> float: ...
    @abc.abstractmethod
    def to_heb(self) -> HebrewDate: ...
    def __hash__(self) -> int: ...
    def __add__(self, other: float) -> BaseDate: ...
    @overload
    def __sub__(self, other: float) -> BaseDate: ...
    @overload
    def __sub__(self, other: BaseDate) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def weekday(self) -> int: ...
    def isoweekday(self) -> int: ...
    def shabbos(self) -> Self: ...
    def fast_day(self, hebrew: bool = False) -> str | None: ...
    def festival(
        self, israel: bool = False, hebrew: bool = False, include_working_days: bool = True, prefix_day: bool = False
    ) -> str | None: ...
    def holiday(self, israel: bool = False, hebrew: bool = False, prefix_day: bool = False) -> str | None: ...

class CalendarDateMixin:
    year: int
    month: int
    day: int
    def __init__(self, year: int, month: int, day: int, jd: float | None = None) -> None: ...
    def __iter__(self) -> Generator[int]: ...
    def tuple(self) -> tuple[int, int, int]: ...
    def dict(self) -> _DateDict: ...
    def replace(self, year: int | None = None, month: int | None = None, day: int | None = None) -> Self: ...

class JulianDay(BaseDate):
    day: float
    def __init__(self, day: float) -> None: ...
    @property
    def jd(self) -> float: ...
    @staticmethod
    def from_pydate(pydate: datetime.date) -> JulianDay: ...
    @staticmethod
    def today() -> JulianDay: ...
    def to_greg(self) -> GregorianDate: ...
    def to_heb(self) -> HebrewDate: ...
    def to_pydate(self) -> datetime.date: ...

class GregorianDate(BaseDate, CalendarDateMixin):
    def __init__(self, year: int, month: int, day: int, jd: float | None = None) -> None: ...
    def __format__(self, fmt: str) -> str: ...
    def strftime(self, fmt: str) -> str: ...
    @property
    def jd(self) -> float: ...
    @classmethod
    def from_pydate(cls, pydate: datetime.date) -> Self: ...
    @staticmethod
    def today() -> GregorianDate: ...
    def is_leap(self) -> bool: ...
    def to_jd(self) -> JulianDay: ...
    def to_heb(self) -> HebrewDate: ...
    def to_pydate(self) -> datetime.date: ...

class HebrewDate(BaseDate, CalendarDateMixin):
    def __init__(self, year: int, month: int, day: int, jd: float | None = None) -> None: ...
    def __format__(self, fmt: str) -> str: ...
    @property
    def jd(self) -> float: ...
    @staticmethod
    def from_pydate(pydate: datetime.date) -> HebrewDate: ...
    @staticmethod
    def today() -> HebrewDate: ...
    def to_jd(self) -> JulianDay: ...
    def to_greg(self) -> GregorianDate: ...
    def to_pydate(self) -> datetime.date: ...
    def to_heb(self) -> HebrewDate: ...
    def month_name(self, hebrew: bool = False) -> str: ...
    def hebrew_day(self, withgershayim: bool = True) -> str: ...
    def hebrew_year(self, thousands: bool = False, withgershayim: bool = True) -> str: ...
    def hebrew_date_string(self, thousands: bool = False) -> str: ...
    def add(
        self, years: int = 0, months: int = 0, days: int = 0, adar1: bool | None = False, rounding: Rounding = Rounding.NEXT_DAY
    ) -> HebrewDate: ...
    def subtract(
        self, years: int = 0, months: int = 0, days: int = 0, adar1: bool | None = False, rounding: Rounding = Rounding.NEXT_DAY
    ) -> HebrewDate: ...



from _typeshed import Incomplete
from typing import Final

__version__: Final[str]

class Percentage(float): ...

class Validator:
    def __call__(self, x): ...
    def normalize(self, x): ...
    def normalizeTest(self, x): ...

class _isAnything(Validator):
    def test(self, x): ...

class _isNothing(Validator):
    def test(self, x): ...

class _isBoolean(Validator):
    def test(self, x): ...
    def normalize(self, x): ...

class _isString(Validator):
    def test(self, x): ...

class _isCodec(Validator):
    def test(self, x): ...

class _isNumber(Validator):
    def test(self, x): ...
    def normalize(self, x): ...

class _isInt(Validator):
    def test(self, x): ...
    def normalize(self, x): ...

class _isNumberOrNone(_isNumber):
    def test(self, x): ...
    def normalize(self, x): ...

class _isListOfNumbersOrNone(Validator):
    def test(self, x): ...

class isNumberInRange(_isNumber):
    min: Incomplete
    max: Incomplete
    def __init__(self, min, max) -> None: ...
    def test(self, x): ...

class _isListOfShapes(Validator):
    def test(self, x): ...

class _isListOfStringsOrNone(Validator):
    def test(self, x): ...

class _isTransform(Validator):
    def test(self, x): ...

class _isColor(Validator):
    def test(self, x): ...

class _isColorOrNone(Validator):
    def test(self, x): ...

class _isNormalDate(Validator):
    def test(self, x): ...
    def normalize(self, x): ...

class _isValidChild(Validator):
    def test(self, x): ...

class _isValidChildOrNone(_isValidChild):
    def test(self, x): ...

class _isCallable(Validator):
    def test(self, x): ...

class OneOf(Validator):
    def __init__(self, enum, *args) -> None: ...
    def test(self, x): ...

class SequenceOf(Validator):
    def __init__(self, elemTest, name=None, emptyOK: int = 1, NoneOK: int = 0, lo: int = 0, hi: int = 2147483647) -> None: ...
    def test(self, x): ...

class EitherOr(Validator):
    def __init__(self, tests, name=None) -> None: ...
    def test(self, x): ...

class NoneOr(EitherOr):
    def test(self, x): ...

class NotSetOr(EitherOr):
    def test(self, x): ...
    @staticmethod
    def conditionalValue(v, a): ...

class _isNotSet(Validator):
    def test(self, x): ...

class Auto(Validator):
    def __init__(self, **kw) -> None: ...
    def test(self, x): ...

class AutoOr(EitherOr):
    def test(self, x): ...

class isInstanceOf(Validator):
    def __init__(self, klass=None) -> None: ...
    def test(self, x): ...

class isSubclassOf(Validator):
    def __init__(self, klass=None) -> None: ...
    def test(self, x): ...

class matchesPattern(Validator):
    def __init__(self, pattern) -> None: ...
    def test(self, x): ...

class DerivedValue:
    def getValue(self, renderer, attr) -> None: ...

class Inherit(DerivedValue):
    def getValue(self, renderer, attr): ...

inherit: Incomplete

class NumericAlign(str):
    def __new__(cls, dp: str = ".", dpLen: int = 0): ...

isAuto: Auto
isBoolean: _isBoolean
isString: _isString
isCodec: _isCodec
isNumber: _isNumber
isInt: _isInt
isNoneOrInt: NoneOr
isNumberOrNone: _isNumberOrNone
isTextAnchor: OneOf
isListOfNumbers: SequenceOf
isListOfNoneOrNumber: SequenceOf
isListOfListOfNoneOrNumber: SequenceOf
isListOfNumbersOrNone: _isListOfNumbersOrNone
isListOfShapes: _isListOfShapes
isListOfStrings: SequenceOf
isListOfStringsOrNone: _isListOfStringsOrNone
isTransform: _isTransform
isColor: _isColor
isListOfColors: SequenceOf
isColorOrNone: _isColorOrNone
isShape: _isValidChild
isValidChild: _isValidChild
isNoneOrShape: _isValidChildOrNone
isValidChildOrNone: _isValidChildOrNone
isAnything: _isAnything
isNothing: _isNothing
isXYCoord: SequenceOf
isBoxAnchor: OneOf
isNoneOrString: NoneOr
isNoneOrListOfNoneOrStrings: SequenceOf
isListOfNoneOrString: SequenceOf
isNoneOrListOfNoneOrNumbers: SequenceOf
isCallable: _isCallable
isNoneOrCallable: NoneOr
isStringOrCallable: EitherOr
isStringOrCallableOrNone: NoneOr
isStringOrNone: NoneOr
isNormalDate: _isNormalDate
isNotSet: _isNotSet



from email.message import Message
from typing import Any, Final, overload
from urllib.error import HTTPError as _HTTPError

class HTTPError(Exception):
    status_code: int
    reason: str
    body: bytes
    headers: Message
    @overload
    def __init__(self, status_code: int, reason: str, body: bytes, headers: Message, /) -> None: ...
    @overload
    def __init__(self, http_error: _HTTPError, /) -> None: ...
    def __reduce__(self) -> tuple[type[HTTPError], tuple[int, str, bytes, Message]]: ...
    @property
    def to_dict(self) -> dict[str, Any]: ...  # dict of response error from the API

class BadRequestsError(HTTPError): ...
class UnauthorizedError(HTTPError): ...
class ForbiddenError(HTTPError): ...
class NotFoundError(HTTPError): ...
class MethodNotAllowedError(HTTPError): ...
class PayloadTooLargeError(HTTPError): ...
class UnsupportedMediaTypeError(HTTPError): ...
class TooManyRequestsError(HTTPError): ...
class InternalServerError(HTTPError): ...
class ServiceUnavailableError(HTTPError): ...
class GatewayTimeoutError(HTTPError): ...

err_dict: Final[dict[int, type[HTTPError]]]

def handle_error(error: _HTTPError) -> HTTPError: ...
